"""
Copyright (c) 2024 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
from __future__ import annotations
import os
from typing import TYPE_CHECKING
from pathlib import Path

from proton.utils.environment import VPNExecutionEnvironment
from proton.vpn.session.utils import RefreshCalculator, rest_api_request
from proton.vpn.core.cache_handler import CacheHandler

if TYPE_CHECKING:
    from proton.vpn.session.api import VPNSession


REFRESH_INTERVAL = 2 * 60 * 60  # 2 hours
FF_ENV_VAR = "PROTON_VPN_FEATURE_FLAG_{flag}"

DEFAULT = {
    "toggles": [
        {
            "name": "BinaryServerStatus",
            "enabled": False,
            "impressionData": False,
            "variant": {
                "name": "disabled",
                "enabled": False
            }
        },
    ],
    "ExpirationTime": 0
}


class FeatureFlags:  # pylint: disable=too-few-public-methods
    """Contains a record of available features."""
    def __init__(self, api_data: dict):
        self._api_data = api_data
        self._expiration_time = api_data.get(
            "ExpirationTime",
            RefreshCalculator.get_expiration_time(
                refresh_interval=REFRESH_INTERVAL
            )
        )

    def get(self, feature_flag_name: str) -> bool:
        """Get a feature flag by its name.
        Always returns `False` if the feature flag is not found.
        """

        # Allow environment variable override for feature flags
        # this makes it easy to enable a feature flag for testing purposes
        if self._search_for_env_var(feature_flag_name):
            return True

        return self._search_for_feature_flag(feature_flag_name)

    def _search_for_env_var(self, feature_name: str) -> bool:
        env_var = FF_ENV_VAR.format(flag=feature_name)
        return env_var in os.environ

    def _search_for_feature_flag(self, feature_name: str) -> bool:
        feature_flag_dict = {}

        for feature in self._api_data.get("toggles", {}):
            if feature["name"] == feature_name:
                feature_flag_dict = feature
                break

        return feature_flag_dict.get("enabled", False)

    @property
    def is_expired(self) -> bool:
        """Returns if data has expired"""
        return RefreshCalculator.get_is_expired(self._expiration_time)

    @property
    def seconds_until_expiration(self) -> int:
        """Returns amount of seconds until it expires."""
        return RefreshCalculator.get_seconds_until_expiration(self._expiration_time)

    @staticmethod
    def get_refresh_interval_in_seconds() -> int:
        """Returns refresh interval in seconds."""
        return RefreshCalculator(REFRESH_INTERVAL).get_refresh_interval_in_seconds()

    @staticmethod
    def default() -> FeatureFlags:
        """Returns a feature object with default values"""
        return FeatureFlags(DEFAULT)


class FeatureFlagsFetcher:
    """Fetches and caches features from Proton's REST API."""
    ROUTE = "/feature/v2/frontend"
    CACHE_PATH = Path(VPNExecutionEnvironment().path_cache) / "features.json"

    def __init__(
        self, session: "VPNSession",
        refresh_calculator: RefreshCalculator = None,
        cache_handler: CacheHandler = None
    ):
        """
        :param session: session used to retrieve the client configuration.
        """
        self._features = None
        self._session = session
        self._refresh_calculator = refresh_calculator or RefreshCalculator
        self._cache_file = cache_handler or CacheHandler(self.CACHE_PATH)

    def clear_cache(self):
        """Discards the cache, if existing."""
        self._features = None
        self._cache_file.remove()

    async def fetch(self) -> FeatureFlags:
        """
        Fetches the client configuration from the REST API.
        :returns: the fetched client configuration.
        """
        response = await rest_api_request(
            self._session,
            self.ROUTE,
        )
        response["ExpirationTime"] = self._refresh_calculator\
            .get_expiration_time(refresh_interval=REFRESH_INTERVAL)
        self._cache_file.save(response)
        self._features = FeatureFlags(response)
        return self._features

    def load_from_cache(self) -> FeatureFlags:
        """
        Loads the client configuration from persistence.
        :returns: the persisted client configuration. If no persistence
            was found then the default client configuration is returned.
        """
        cache = self._cache_file.load()
        self._features = FeatureFlags(cache) if cache else FeatureFlags.default()
        return self._features



"""
Copyright (c) 2023 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
from __future__ import annotations
from dataclasses import dataclass
from typing import Optional
from proton.vpn.session.utils import Serializable

# pylint: disable=invalid-name


@dataclass
class VPNLocation(Serializable):
    """Data about the physical location the VPN client runs from."""
    IP: str
    Country: str
    ISP: str
    Long: Optional[float]
    Lat: Optional[float]

    @staticmethod
    def _deserialize(dict_data: dict) -> VPNLocation:
        """
        Builds a Location object from a dict containing the parsed
        JSON response returned by the API.
        """
        return VPNLocation(
            IP=dict_data["IP"],
            Country=dict_data["Country"],
            ISP=dict_data["ISP"],
            Long=dict_data.get("Long"),
            Lat=dict_data.get("Lat"),
        )



"""
VPN connection interface.


Copyright (c) 2023 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""

from __future__ import annotations

import asyncio
import sys
from abc import ABC, abstractmethod
from typing import Callable, List

from proton.loader import Loader

from proton.vpn.connection.events import Event, EventContext
from proton.vpn.connection.interfaces import VPNServer, Settings, VPNCredentials
from proton.vpn.connection.persistence import ConnectionPersistence, ConnectionParameters
from proton.vpn.connection.publisher import Publisher
from proton.vpn.connection import states, events


# pylint: disable=too-many-instance-attributes
class VPNConnection(ABC):
    """
    Defines the interface to create a new VPN connection.

    It's the base class for any VPN connection implementation.
    """

    # Class attrs to be set by subclasses.
    backend = None
    protocol = None

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        server: VPNServer,
        credentials: VPNCredentials,
        settings: Settings,
        connection_id: str = None,
        connection_persistence: ConnectionPersistence = None,
        publisher: Publisher = None,
    ):
        """Initialize a VPNConnection object.

        :param server: VPN server to connect to.
        :param credentials: credentials used to authenticate to the VPN server.
        :param settings: Settings to be used when establishing the VPN connection.
            This parameter is optional. When it's not specified the default settings
            will be used instead.
        :param connection_id: unique ID of the existing connection.
            This parameter is optional. It should be specified only if this instance
            maps to an already existing network connection.
        :param connection_persistence: Connection persistence implementation.
            This parameter is optional. When not specified, the default connection
            persistence implementation will be used instead.
        :param publisher: Publisher implementation. This parameter is optional. Pass it
            only if you know what you are doing.
        """
        self._vpnserver = server
        self._vpncredentials = credentials
        self._settings = settings

        self._connection_persistence = connection_persistence or ConnectionPersistence()
        self._publisher = publisher or Publisher()

        if connection_id:
            self._unique_id = connection_id
            self.initial_state = self._initialize_persisted_connection(
                connection_id
            )
        else:
            self._unique_id = None
            self.initial_state = states.Disconnected(
                states.StateContext(
                    event=events.Initialized(EventContext(connection=self)),
                    connection=self
                )
            )

    @abstractmethod
    def _initialize_persisted_connection(self, connection_id: str) -> states.State:
        """
        Initializes the state of this instance of VPN connection according
        to previously persisted connection parameters and returns its current state.
        Needs to be provided by the VPN connection implementation.
        """

    @abstractmethod
    async def start(self):
        """
        Starts the VPN connection.

        This method returns as soon as the connection has been started, but
        it doesn't wait for the connection to be fully established.
        """

    @abstractmethod
    async def stop(self):
        """Stops the VPN connection."""

    @property
    def are_feature_updates_applied_when_active(self) -> bool:
        """
        Returns whether the connection features updates are applied on the fly
        while the connection is already active, without restarting the connection.
        """
        return False

    async def update_credentials(self, credentials: VPNCredentials):
        """
        Updates the connection credentials.
        """
        self._vpncredentials = credentials
        # Note that VPN connection implementations can extend this method to send
        # the new credentials to the back-end. That's why  this method is left async.

    async def update_settings(self, settings: Settings):
        """
        Updates the connection settings.
        """
        self._settings = settings
        # Note that VPN connection implementations can extend this method to send
        # the new settings to the back-end. That's why this method is left async.

    def register(self, subscriber: Callable[[Event], None]):
        """
        Registers a subscriber to be notified whenever a new connection event happens.

        The subscriber will be called passing the connection event as argument.
        """
        self._publisher.register(subscriber)

    def unregister(self, subscriber: Callable[[Event], None]):
        """Unregister a previously registered connection events subscriber."""
        self._publisher.unregister(subscriber)

    def _notify_subscribers(self, event: Event):
        """Notifies all subscribers of a connection event.

        Subscribers are called passing the connection event as argument.

        This is a utility method that VPN connection implementations can use to notify
        subscribers when a new connection event happens.

        :param event: the event to be notified to subscribers.
        """
        self._publisher.notify(event=event)

    @staticmethod
    def create(server: VPNServer, credentials: VPNCredentials, settings: Settings = None,
               protocol: str = None, backend: str = None):  # pylint: disable=too-many-arguments
        """
        Creates a new VPN connection object. Note the VPN connection won't be initiated. For that
        to happen, see the `start` method.

        :param server: VPN server to connect to.
        :param credentials: Credentials used to authenticate to the VPN server.
        :param settings: VPN settings used to create the connection.
        :param protocol: protocol to connect with. If None, the default protocol will be used.
        :param backend: Name of the class implementing the VPNConnection interface.
            If None, the default implementation will be used.
        """
        backend = Loader.get("backend", class_name=backend)
        protocol = protocol.lower() if protocol else None
        protocol_class = backend.factory(protocol)
        return protocol_class(server, credentials, settings)

    @property
    def server(self) -> VPNServer:
        """Returns the VPN server of this VPN connection."""
        return self._vpnserver

    @property
    def server_id(self) -> str:
        """Returns the VPN server ID of this VPN connection."""
        return self._vpnserver.server_id

    @property
    def server_name(self) -> str:
        """Returns the VPN server name of this VPN connection."""
        return self._vpnserver.server_name

    @property
    def server_ip(self) -> str:
        """Returns the VPN server IP of this VPN connection."""
        return self._vpnserver.server_ip

    @property
    def server_domain(self) -> str:
        """Returns the VPN server domain of this VPN connection."""
        return self._vpnserver.domain

    @property
    def settings(self) -> Settings:
        """ Current settings of the connection :
            Some settings can be changed on the fly and are RW :
            netshield level, kill switch enabled/disabled, split tunneling,
            VPN accelerator, custom DNS.
            Other settings are RO and cannot be changed once the connection
            is instantiated: VPN protocol.
        """
        return self._settings

    @classmethod
    @abstractmethod
    def _get_priority(cls) -> int:
        """
        Priority of the VPN connection implementation.

        To be implemented by subclasses.

        When no backend is specified when creating a VPN connection instance
        with `VPNConnection.create`, the VPN connection implementation is
        chosen based on the priority value returned by this method.

        The lower the value, the more priority it has.

        Ideally, the returned priority value should not be hardcoded but
        calculated based on the environment. For example, a VPN connection
        implementation using NetworkManager could return a high priority
        when the NetworkManager service is running or a low priority when it's
        not.
        """

    @classmethod
    @abstractmethod
    def _validate(cls) -> bool:
        """
        Determines whether the VPN connection implementation is valid or not.
        To be implemented by subclasses.

        If this method returns `False` then the VPN connection implementation
        will be skipped when creating a VPN connection instance with
        `VPNConnection.create`.

        :return: `True` if the implementation is valid or `False` otherwise.
        """

    async def add_persistence(self):
        """
        Stores the connection parameters to disk.

        The connection parameters (e.g. backend, protocol, connection ID,
        server name) are stored to disk so that they can be loaded again
        after an unexpected crash.
        """
        params = ConnectionParameters(
            connection_id=self._unique_id,
            backend=type(self).backend,
            protocol=type(self).protocol,
            server=self.server
        )
        loop = asyncio.get_running_loop()
        await loop.run_in_executor(None, self._connection_persistence.save, params)

    async def remove_persistence(self):
        """
        Works in the opposite way of add_persistence. It removes the
        persistence file. This is used in conjunction with down, since if the
        connection is turned down, we don't want to keep any persistence files.
        """
        loop = asyncio.get_running_loop()
        await loop.run_in_executor(None, self._connection_persistence.remove)

    def _get_user_pass(self, apply_feature_flags=False):
        """*For developers*

        :param apply_feature_flags: if feature flags are to be suffixed to username

        In case of non-certificate based authentication, username and password need
        to be provided for authentication. In such cases, the username can be optionally
        suffixed with different options, of which are fetched from `self._settings`

        Usage:

        .. code-block::

            from proton.vpn.connection import VPNConnection

            class CustomBackend(VPNConnection):
                backend = "custom_backend"

                ...

                def _setup(self):
                    if not use_ceritificate:
                        # In this case, the username will have suffixes added given
                        # that any of the them are set in `self._settings`
                        user, pass = self._get_user_pass()

                        # Then add the username and password to the configurations

        """
        user_data = self._vpncredentials.userpass_credentials
        username = user_data.username
        if apply_feature_flags:
            flags = self._get_feature_flags()
            username = "+".join([username] + flags)  # each flag must be preceded by "+"

        return username, user_data.password

    def _get_feature_flags(self) -> List[str]:
        """
        Creates a list of feature flags that are fetched from `self._settings`.
        These feature flags are used to suffix them to a username, to trigger server-side
        specific behavior.
        """
        list_flags = []
        label = self._vpnserver.label

        if sys.platform.startswith("linux"):
            list_flags.append("pl")
        elif sys.platform.startswith("win32") or sys.platform.startswith("cygwin"):
            list_flags.append("pw")
        elif sys.platform.startswith("darwin"):
            list_flags.append("pm")

        # This is used to ensure that the provided IP matches the one
        # from the exit IP.
        if label:
            list_flags.append(f"b:{label}")

        if self._settings is None:
            return list_flags

        enable_ipv6_support = self._vpnserver.has_ipv6_support and self._settings.ipv6

        if enable_ipv6_support:
            list_flags.append("6")

        features = self._settings.features

        # We only need to add feature flags if there are any
        if features:
            list_flags.append(f"f{features.netshield}")

            if not features.vpn_accelerator:
                list_flags.append("nst")
            if features.port_forwarding:
                list_flags.append("pmp")
            if features.moderate_nat:
                list_flags.append("nr")

        return list_flags



"""
Copyright (c) 2023 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
from __future__ import annotations
from typing import List
from dataclasses import dataclass
from proton.vpn.session.utils import Serializable

# pylint: disable=invalid-name


@dataclass
class VPNInfo(Serializable):  # pylint: disable=too-many-instance-attributes
    """ Same object structure as the one coming from the API"""
    ExpirationTime: int
    Name: str
    Password: str
    GroupID: str
    Status: int
    PlanName: str
    PlanTitle: str
    MaxTier: int
    """ Maximum tier value that this account can vpn connect to """
    MaxConnect: int
    """ Maximum number of simultaneous session on the infrastructure"""
    Groups: List[str]
    """ List of groups that this account belongs to """
    NeedConnectionAllocation: bool

    @staticmethod
    def _deserialize(dict_data: dict) -> VPNInfo:
        return VPNInfo(
            ExpirationTime=dict_data["ExpirationTime"],
            Name=dict_data["Name"],
            Password=dict_data["Password"],
            GroupID=dict_data["GroupID"],
            Status=dict_data["Status"],
            PlanName=dict_data["PlanName"],
            PlanTitle=dict_data["PlanTitle"],
            MaxTier=dict_data["MaxTier"],
            MaxConnect=dict_data["MaxConnect"],
            Groups=dict_data["Groups"],
            NeedConnectionAllocation=dict_data["NeedConnectionAllocation"]
        )


@dataclass
class VPNSettings(Serializable):  # pylint: disable=too-many-instance-attributes
    """ Same object structure as the one coming from the API"""
    VPN: VPNInfo
    Services: int
    Subscribed: int
    Delinquent: int
    """ Encode the deliquent status of the account """
    HasPaymentMethod: int
    Credit: int
    Currency: str
    Warnings: List[str]

    @staticmethod
    def _deserialize(dict_data: dict) -> VPNSettings:
        return VPNSettings(
            VPN=VPNInfo.from_dict(dict_data["VPN"]),
            Services=dict_data["Services"],
            Subscribed=dict_data["Subscribed"],
            Delinquent=dict_data["Delinquent"],
            HasPaymentMethod=dict_data["HasPaymentMethod"],
            Credit=dict_data["Credit"],
            Currency=dict_data["Currency"],
            Warnings=dict_data["Warnings"]
        )



"""
Copyright (c) 2023 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
from __future__ import annotations

import base64
import random

from typing import Optional
from proton.vpn.session.certificates import Certificate
from proton.vpn.session.dataclasses import VPNCertificate
from proton.vpn.session.exceptions import VPNCertificateFingerprintError
from proton.vpn.session.key_mgr import KeyHandler
from proton.vpn import logging


logger = logging.getLogger(__name__)


class VPNSecrets:
    """ Asymmetric crypto secrets generated locally by the client to :

        - connect to the VPN service
        - ask for a certificate to the API with the corresponding public key.
    """
    def __init__(self, ed25519_privatekey: Optional[str] = None):
        self._key_handler = (
            KeyHandler(base64.b64decode(ed25519_privatekey))
            if ed25519_privatekey
            else KeyHandler()
        )

    def get_ed5519_sk_pem(self, password: Optional[bytes] = None):
        """
        Returns the ed5519 private key in pem format,
        and encrypted if a password was passed.
        """
        return self._key_handler.get_ed25519_sk_pem(password)

    @property
    def wireguard_privatekey(self) -> str:
        """Wireguard private key encoded in base64.
            To be added locally by the user. The API route is not providing it.
        """
        return self._key_handler.x25519_sk_str

    @property
    def ed25519_privatekey(self) -> str:
        """Private key in ed25519 base64 format. used to check fingerprints"""
        return self._key_handler.ed25519_sk_str

    @property
    def ed25519_pk_pem(self) -> str:  # pylint: disable=missing-function-docstring
        return self._key_handler.ed25519_pk_pem

    @property
    def proton_fingerprint_from_x25519_pk(self):  # pylint: disable=missing-function-docstring
        return self._key_handler.get_proton_fingerprint_from_x25519_pk(
            self._key_handler.x25519_pk_bytes
        )

    @staticmethod
    def from_dict(dict_data: dict):  # pylint: disable=missing-function-docstring
        return VPNSecrets(dict_data["ed25519_privatekey"])

    def to_dict(self):  # pylint: disable=missing-function-docstring
        return {
            "ed25519_privatekey": self.ed25519_privatekey
        }


class VPNPubkeyCredentials:
    """ Class responsible to hold vpn public key API RAW certificates and
        its associated private key for authentication.
    """
    MINIMUM_VALIDITY_PERIOD_IN_SECS = 300
    # FIXME: We were asked to increase the certification duration  # pylint: disable=fixme
    #  to 7 days due to certificate refresh issues, until a proper fix is put in place.
    #  It should be reverted to 1 day.
    REFRESH_INTERVAL = 60 * 60 * 24 * 7
    REFRESH_RANDOMNESS = 0.22  # +/- 22%

    def __init__(self, api_certificate: VPNCertificate, secrets: VPNSecrets, strict: bool = True):
        self._api_certificate = api_certificate
        self._secrets = secrets

        self._certificate_obj = self._build_certificate(
            api_certificate,
            secrets,
            strict
        )

    @classmethod
    def _generate_random_component(cls):
        # 1 +/- 0.22*random  # nosec B311
        return 1 + cls.REFRESH_RANDOMNESS * (2 * random.random() - 1)  # nosec B311 # noqa: E501 # pylint: disable=line-too-long # nosemgrep: gitlab.bandit.B311

    @classmethod
    def get_refresh_interval_in_seconds(cls):  # pylint: disable=missing-function-docstring
        return cls.REFRESH_INTERVAL * cls._generate_random_component()

    def _build_certificate(self, api_certificate, secrets, strict):
        fingerprint_from_secrets = secrets.proton_fingerprint_from_x25519_pk

        # Get fingerprint from Certificate public key
        certificate = Certificate(cert_pem=api_certificate.Certificate)
        fingerprint_from_certificate = certificate.proton_fingerprint

        # Refuse to store unmatching fingerprints when strict equal True
        if strict:
            if fingerprint_from_secrets != fingerprint_from_certificate:
                raise VPNCertificateFingerprintError

        return Certificate(cert_pem=api_certificate.Certificate)

    def get_ed25519_sk_pem(self, password: Optional[bytes] = None):
        """
        Returns the ed5519 private key in pem format,
        and encrypted if a password was passed.
        """
        return self._secrets.get_ed5519_sk_pem(password)

    @property
    def certificate_pem(self) -> str:
        """ X509 client certificate in PEM format, can be used
            to connect for client based authentication to the local agent

            :raises VPNCertificateNotAvailableError: : certificate cannot be found
                :class:`VPNSession` must be populated with :meth:`VPNSession.refresh`.
            :return: :class:`api_data.VPNCertificate.Certificate`
        """
        self._log_if_certificate_requires_to_be_refreshed_but_is_not_expired()
        return self._certificate_obj.get_as_pem()

    @property
    def openvpn_private_key(self) -> str:
        """ Get OpenVPN private key in pem format, directly usable in an
            OpenVPN configuration file.
        """
        self._log_if_certificate_requires_to_be_refreshed_but_is_not_expired()
        return self._secrets.get_ed5519_sk_pem()

    @property
    def wg_private_key(self) -> str:
        """ Get Wireguard private key in base64 format,
            directly usable in a wireguard configuration file. This key
            is tied to the Proton :class:`VPNCertCredentials` by its
            corresponding API certificate.

            :return: :class:`api_data.VPNSecrets.wireguard_privatekey`: Wireguard private key
                in base64 format.
        """
        self._log_if_certificate_requires_to_be_refreshed_but_is_not_expired()
        return self._secrets.wireguard_privatekey

    @property
    def ed_255519_private_key(self) -> str:  # pylint: disable=missing-function-docstring
        return self._secrets.ed25519_privatekey

    @property
    def certificate_validity_remaining(self) -> Optional[float]:
        """ remaining time the certificate is valid, in seconds.

            - < 0 : certificate is not valid anymore
            -  None we don't have a certificate.
        """
        return self._certificate_obj.validity_period

    @property
    def remaining_time_to_next_refresh(self) -> int:
        """Returns a timestamp of when the next refresh should be done."""
        return self._api_certificate.remaining_time_to_next_refresh

    @property
    def proton_extensions(self):  # pylint: disable=missing-function-docstring
        return self._certificate_obj.proton_extensions

    @property
    def certificate_duration(self) -> Optional[float]:
        """ certificate range in seconds, even if not valid anymore.

            - return `None` if we don't have a certificate
        """
        return self._certificate_obj.duration.total_seconds()

    def _log_if_certificate_requires_to_be_refreshed_but_is_not_expired(self):
        if (
            self._certificate_obj.validity_period
            <= VPNPubkeyCredentials.MINIMUM_VALIDITY_PERIOD_IN_SECS
        ):
            logger.warning(
                msg="Current certificate will expire.",
                category="CREDENTIALS",
                subcategory="CERTIFICATE", event="REQUIRE_REFRESH"
            )



"""
Utility module to do TCP connection checks.
"""
import asyncio
import ipaddress
import logging
import socket
from asyncio import as_completed
from contextlib import closing
from typing import List, Union

logger = logging.getLogger(__name__)

DEFAULT_TIMEOUT = 5


def _get_address_family(
        ip_address: Union[ipaddress.IPv4Address, ipaddress.IPv6Address]
) -> socket.AddressFamily:  # pylint: disable=no-member
    if isinstance(ip_address, ipaddress.IPv4Address):
        return socket.AF_INET
    if isinstance(ip_address, ipaddress.IPv6Address):
        return socket.AF_INET6
    raise TypeError(f"Invalid IP address: {ip_address}")


def is_port_reachable(
        ip_address: str, port: str, timeout_in_secs: int = DEFAULT_TIMEOUT
) -> bool:
    """
    Checks if the specified IP address is reachable by opening a TCP socket
    to the specified port.

    :param ip_address: IP address to connect to.
    :param port: TCP port to connect to.
    :param timeout_in_secs: Optional connection timeout.
    :returns: True if a socket could be opened to the specified address/port,
        or False otherwise.
    """
    ip_address = ipaddress.ip_address(ip_address)
    address_family = _get_address_family(ip_address)

    logger.debug("Checking %s:%s...", ip_address, port)
    with closing(socket.socket(address_family, socket.SOCK_STREAM)) as sock:
        sock.settimeout(timeout_in_secs)
        socket_result = sock.connect_ex((f"{ip_address}", port))
        logger.debug("%s:%s => %s", ip_address, port, socket_result)
        return socket_result == 0


async def is_any_port_reachable(
        ip_address: str, ports: List[str], timeout: int = DEFAULT_TIMEOUT
) -> bool:
    """
    Checks if the specified IP address is reachable by opening a TCP socket
    to any of the specified ports.

    All ports are probed in parallel. As soon as a socket is opened to
    one of them, connectivity is reported by returning True. If all
    the connections to the different ports fail, then False is returned.

    :param ip_address: IP address to connect to.
    :param port: TCP ports to try to connect to.
    :param timeout_in_secs: Optional connection timeout.
    :returns: True if a socket could be opened to the specified address/ports,
        or False otherwise.
    """
    loop = asyncio.get_running_loop()

    async def _is_port_reachable(port):
        return await loop.run_in_executor(None, is_port_reachable, ip_address, port, timeout)

    tasks = [
        asyncio.create_task(_is_port_reachable(port))
        for port in ports
    ]
    for task in as_completed(tasks):
        try:
            return await task
        except Exception:  # pylint: disable=broad-except
            return False



"""
Copyright (c) 2025 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
from typing import Optional, Protocol


class UserInteraction(Protocol):
    """
    Provides user interaction to the Fido2 client.

    The interface currently follows the fido2.client.UserInteraction interface,
    adding some methods to it.
    """

    def prompt_up(self) -> None:
        """Called when the authenticator is awaiting a user presence check."""

    def request_pin(
        self, *args, **kwargs
    ) -> Optional[str]:
        """Called when the client requires a PIN from the user.

        Should return a PIN, or None/Empty to cancel."""

    def request_uv(self, *args, **kwargs) -> bool:
        """Called when the client is about to request UV from the user.

        Should return True if allowed, or False to cancel."""

    def request_key_selection(self) -> None:
        """Called when multiple keys are found and the user needs to select one
        by touching it."""



"""
This module manages the Proton VPN general settings.


Copyright (c) 2023 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
from proton.vpn.core.settings.settings import Settings, SettingsPersistence
from proton.vpn.core.settings.split_tunneling import SplitTunneling, \
    SplitTunnelingConfig, SplitTunnelingMode
from proton.vpn.core.settings.features import NetShield
from proton.vpn.core.settings.custom_dns import CustomDNSEntry

__all__ = [
    "Settings", "SettingsPersistence", "NetShield",
    "CustomDNSEntry", "SplitTunneling", "SplitTunnelingConfig", "SplitTunnelingMode"
]



"""
Implementation of the Publisher/Subscriber used to signal VPN connection
state changes.


Copyright (c) 2023 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
import asyncio
import inspect
from typing import Callable, List, Optional

from proton.vpn import logging

logger = logging.getLogger(__name__)


class Publisher:
    """Simple generic implementation of the publish-subscribe pattern."""

    def __init__(self, subscribers: Optional[List[Callable]] = None):
        self._subscribers = subscribers or []
        self._pending_tasks = set()

    def register(self, subscriber: Callable):
        """
        Registers a subscriber to be notified of new updates.

        The subscribers are not expected to block, as they will be notified
        sequentially, one after the other in the order in which they were
        registered.

        :param subscriber: callback that will be called with the expected
            args/kwargs whenever there is an update.
        :raises ValueError: if the subscriber is not callable.
        """
        if not callable(subscriber):
            raise ValueError(f"Subscriber to register is not callable: {subscriber}")

        if subscriber not in self._subscribers:
            self._subscribers.append(subscriber)

    def unregister(self, subscriber: Callable):
        """
        Unregisters a subscriber.

        :param subscriber: the subscriber to be unregistered.
        """
        if subscriber in self._subscribers:
            self._subscribers.remove(subscriber)

    def notify(self, *args, **kwargs):
        """
        Notifies the subscribers about a new update.

        All subscribers will be called

        Each backend and/or protocol have to call this method whenever the connection
        state changes, so that each subscriber can receive states changes whenever they occur.

            :param connection_status: the current status of the connection
            :type connection_status: ConnectionStateEnum

        """
        for subscriber in self._subscribers:
            try:
                if inspect.iscoroutinefunction(subscriber):
                    notification_task = asyncio.create_task(subscriber(*args, **kwargs))
                    self._pending_tasks.add(notification_task)
                    notification_task.add_done_callback(self._on_notification_task_done)
                else:
                    subscriber(*args, **kwargs)
            except Exception:  # pylint: disable=broad-except
                logger.exception(f"An error occurred notifying subscriber {subscriber}.")

    def _on_notification_task_done(self, task: asyncio.Task):
        self._pending_tasks.discard(task)
        task.result()

    def is_subscriber_registered(self, subscriber: Callable) -> bool:
        """Returns whether a subscriber is registered or not."""
        return subscriber in self._subscribers

    @property
    def number_of_subscribers(self) -> int:
        """Number of currently registered subscribers."""
        return len(self._subscribers)



"""
VPN connection events to react to.


Copyright (c) 2023 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING, Optional, Any

from .enum import StateMachineEventEnum

if TYPE_CHECKING:
    from proton.vpn.connection.vpnconnection import VPNConnection


@dataclass
class ConnectionDetails:
    """Connection details obtained via local agent."""
    device_ip: Optional[str] = None
    device_country: Optional[str] = None
    server_ipv4: Optional[str] = None
    server_ipv6: Optional[str] = None


# pylint: disable=too-few-public-methods

@dataclass
class EventContext:
    """
    Relevant event context.
    Args:
        connection: the VPN connection object that emitted this event.
        reason: optional backend-dependent data providing more context about the event.
        error: an optional exception to be bubbled up while processing the event.
    """
    connection: "VPNConnection"
    connection_details: Optional[ConnectionDetails] = None
    forwarded_port: Optional[int] = None
    reason: Optional[Any] = None
    error: Optional[Exception] = None


class Event:
    """Base event that all the other events should inherit from."""
    type = None

    def __init__(self, context: EventContext = None):
        if self.type is None:
            raise AttributeError("event attribute not defined")

        self.context = context or EventContext(connection=None)

    def check_for_errors(self):
        """Raises an exception if there is one."""
        if self.context.error:
            raise self.context.error


class Initialized(Event):
    """Event that leads to the initial state."""
    type = StateMachineEventEnum.INITIALIZED


class Up(Event):
    """Signals that the VPN connection should be started."""
    type = StateMachineEventEnum.UP


class Down(Event):
    """Signals that the VPN connection should be stopped."""
    type = StateMachineEventEnum.DOWN


class Connected(Event):
    """Signals that the VPN connection was successfully established."""
    type = StateMachineEventEnum.CONNECTED


class Disconnected(Event):
    """Signals that the VPN connection was successfully disconnected by the user."""
    type = StateMachineEventEnum.DISCONNECTED


class Error(Event):
    """Parent class for events signaling VPN disconnection."""


class DeviceDisconnected(Error):
    """Signals that the VPN connection dropped unintentionally."""
    type = StateMachineEventEnum.DEVICE_DISCONNECTED


class Timeout(Error):
    """Signals that a timeout occurred while trying to establish the VPN
    connection."""
    type = StateMachineEventEnum.TIMEOUT


class AuthDenied(Error):
    """Signals that an authentication denied occurred while trying to establish
    the VPN connection."""
    type = StateMachineEventEnum.AUTH_DENIED


class ExpiredCertificate(Error):
    """Signals that the passed certificate has expired and needs to be refreshed."""
    type = StateMachineEventEnum.CERTIFICATE_EXPIRED


class MaximumSessionsReached(Error):
    """Signals that for the given plan the user has too many devices/sessions connected."""
    type = StateMachineEventEnum.MAXIMUM_SESSIONS_REACHED


class TunnelSetupFailed(Error):
    """Signals that there was an error setting up the VPN tunnel."""
    type = StateMachineEventEnum.TUNNEL_SETUP_FAILED


class UnexpectedError(Error):
    """Signals that an unexpected error occurred."""
    type = StateMachineEventEnum.UNEXPECTED_ERROR


class TwoFARequired(Error):
    """Signals that 2 factor authentication is required."""
    type = StateMachineEventEnum.TWOFA_REQUIRED


_event_types = [
    event_type for event_type in Event.__subclasses__()
    if event_type is not Error  # As error is an abstract class.
]
_event_types.extend(Error.__subclasses__())
EVENT_TYPES = tuple(_event_types)



"""
Copyright (c) 2023 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
from __future__ import annotations
from typing import TYPE_CHECKING
from pathlib import Path
import random
import time

from proton.utils.environment import VPNExecutionEnvironment

from proton.vpn.core.cache_handler import CacheHandler
from proton.vpn.session.exceptions import ClientConfigDecodeError
from proton.vpn.session.utils import rest_api_request
from proton.vpn.session.dataclasses.client_config import ProtocolPorts
from proton.vpn import logging

logger = logging.getLogger(__name__)


if TYPE_CHECKING:
    from proton.vpn.session import VPNSession

DEFAULT_CLIENT_CONFIG = {
    "DefaultPorts": {
        "OpenVPN": {
            "UDP": [80, 51820, 4569, 1194, 5060],
            "TCP": [443, 7770, 8443]
        },
        "WireGuard": {
            "UDP": [443, 88, 1224, 51820, 500, 4500],
            "TCP": [443],
        }
    },
    "HolesIPs": ["62.112.9.168", "104.245.144.186"],
    "ServerRefreshInterval": 10,
    "FeatureFlags": {
        "NetShield": True,
        "GuestHoles": False,
        "ServerRefresh": True,
        "StreamingServicesLogos": True,
        "PortForwarding": True,
        "ModerateNAT": True,
        "SafeMode": False,
        "StartConnectOnBoot": True,
        "PollNotificationAPI": True,
        "VpnAccelerator": True,
        "SmartReconnect": True,
        "PromoCode": False,
        "WireGuardTls": True,
        "Telemetry": True,
        "NetShieldStats": True
    },
    "SmartProtocol": {
        "OpenVPN": True,
        "IKEv2": True,
        "WireGuard": True,
        "WireGuardTCP": True,
        "WireGuardTLS": True
    },
    "RatingSettings": {
        "EligiblePlans": [],
        "SuccessConnections": 3,
        "DaysLastReviewPassed": 100,
        "DaysConnected": 3,
        "DaysFromFirstConnection": 14
    }
}


class ClientConfig:
    """
    General configuration used to connect to VPN servers.
    """

    REFRESH_INTERVAL = 3 * 60 * 60  # 3 hours
    REFRESH_RANDOMNESS = 0.22  # +/- 22%

    def __init__(
        self, openvpn_ports, wireguard_ports, holes_ips,
        server_refresh_interval,
        expiration_time
    ):  # pylint: disable=R0913
        self.openvpn_ports = openvpn_ports
        self.wireguard_ports = wireguard_ports
        self.holes_ips = holes_ips
        self.server_refresh_interval = server_refresh_interval
        self.expiration_time = expiration_time

    @classmethod
    def from_dict(cls, apidata: dict) -> ClientConfig:
        """Creates ClientConfig object from data."""
        try:
            openvpn_ports = apidata["DefaultPorts"]["OpenVPN"]
            wireguard_ports = apidata["DefaultPorts"]["WireGuard"]
            holes_ips = apidata["HolesIPs"]
            server_refresh_interval = apidata["ServerRefreshInterval"]
            expiration_time = float(apidata.get("ExpirationTime", cls.get_expiration_time()))

            return ClientConfig(
                # No need to copy openvpn_ports, OpenVPNPorts takes care of it.
                ProtocolPorts.from_dict(openvpn_ports),
                # No need to copy wireguard_ports, WireGuardPorts takes care of it.
                ProtocolPorts.from_dict(wireguard_ports),
                # We copy the holes_ips list to avoid side effects if it's modified.
                holes_ips.copy(),
                server_refresh_interval,
                expiration_time
            )
        except (KeyError, ValueError) as error:
            raise ClientConfigDecodeError(
                "Error parsing client configuration."
            ) from error

    @staticmethod
    def default() -> ClientConfig:
        """":returns: the default client configuration."""
        return ClientConfig.from_dict(DEFAULT_CLIENT_CONFIG)

    @property
    def is_expired(self) -> bool:
        """Returns if data has expired"""
        current_time = time.time()
        return current_time > self.expiration_time

    @property
    def seconds_until_expiration(self) -> float:
        """
        Amount of seconds left until the client configuration is considered
        outdated and should be fetched again from the REST API.
        """
        seconds_left = self.expiration_time - time.time()
        return seconds_left if seconds_left > 0 else 0

    @classmethod
    def _generate_random_component(cls):
        # 1 +/- 0.22*random  # nosec B311
        return 1 + cls.REFRESH_RANDOMNESS * (2 * random.random() - 1)   # nosec B311 # noqa: E501 # pylint: disable=line-too-long # nosemgrep: gitlab.bandit.B311

    @classmethod
    def get_refresh_interval_in_seconds(cls):  # pylint: disable=missing-function-docstring
        return cls.REFRESH_INTERVAL * cls._generate_random_component()

    @classmethod
    def get_expiration_time(cls, start_time: int = None):  # noqa: E501 pylint: disable=missing-function-docstring
        start_time = start_time if start_time is not None else time.time()
        return start_time + cls.get_refresh_interval_in_seconds()


class ClientConfigFetcher:
    """
    Fetches and caches the client configuration from Proton's REST API.
    """
    ROUTE = "/vpn/v2/clientconfig"
    CACHE_PATH = Path(VPNExecutionEnvironment().path_cache) / "clientconfig.json"

    def __init__(self, session: "VPNSession"):
        """
        :param session: session used to retrieve the client configuration.
        """
        self._session = session
        self._client_config = None
        self._cache_file = CacheHandler(self.CACHE_PATH)

    def clear_cache(self):
        """Discards the cache, if existing."""
        self._client_config = None
        self._cache_file.remove()

    async def fetch(self) -> ClientConfig:
        """
        Fetches the client configuration from the REST API.
        :returns: the fetched client configuration.
        """
        response = await rest_api_request(
            self._session,
            self.ROUTE,
        )
        response["ExpirationTime"] = ClientConfig.get_expiration_time()
        self._cache_file.save(response)
        self._client_config = ClientConfig.from_dict(response)
        return self._client_config

    def load_from_cache(self) -> ClientConfig:
        """
        Loads the client configuration from persistence.
        :returns: the persisted client configuration. If no persistence
            was found, or it's invalid then the default client configuration is returned.

        """
        cache = self._cache_file.load()

        if cache:
            try:
                self._client_config = ClientConfig.from_dict(cache)
            except ClientConfigDecodeError:
                logger.warning(
                    "Client config could not be deserialized. Using defaults.",
                    exc_info=True
                )
                self._client_config = ClientConfig.default()
        else:
            self._client_config = ClientConfig.default()

        return self._client_config



"""
Copyright (c) 2023 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
import base64
import datetime
import enum
import hashlib
import typing
import nacl.bindings
import cryptography.x509
from cryptography.hazmat.primitives.serialization import Encoding, PublicFormat
import cryptography.hazmat.backends


class Asn1BerDecoder:  # pylint: disable=missing-class-docstring

    _TYPE_INTEGER = 0x02
    _TYPE_OCTET_STR = 0x04
    _TYPE_SEQUENCE = 0x10
    _TYPE_SEQUENCE_OF = 0x30

    @classmethod
    def __get_asn1_ber_len(cls, raw: bytes) -> typing.Tuple[int, int]:
        """ returns : tuple (length, position start of data) """
        # byte 0 : data type
        if raw[1] & 0x80 == 0:
            # The short form is a single byte, between 0 and 127.
            return raw[1], 2

        # The long form is at least two bytes long, and has bit 8 of the first byte set to 1.
        # Bits 7-1 of the first byte indicate how many more bytes are in
        # the length field itself.
        # Then the remaining bytes specify the length itself, as a multi-byte integer.
        length_of_length = raw[1] & 0x7f
        data_len = 0
        for b in raw[2:2 + length_of_length]:  # pylint: disable=invalid-name
            data_len = data_len * 256 + b
        return data_len, length_of_length + 2

    @classmethod
    def _transform_value_to_str_no_len_check(cls, raw: bytes) -> typing.Tuple[str, int]:
        """ returns : tuple (decoded string, total length) """
        if raw[0] != cls._TYPE_OCTET_STR:
            raise ValueError(f"Not a string : {raw}")
        data_len, pos_data = cls.__get_asn1_ber_len(raw)
        return raw[pos_data:pos_data + data_len].decode("ascii"), (pos_data + data_len)

    @classmethod
    def transform_value_to_str(cls, raw: bytes) -> str:  # noqa: E501 pylint: disable=missing-function-docstring
        data, total_len = cls._transform_value_to_str_no_len_check(raw)
        if total_len != len(raw):
            raise ValueError(
                F"wrong extension length : {raw} , found {total_len}, expected {len(raw)}"
            )
        return data

    @classmethod
    def _transform_value_to_int_no_len_check(cls, raw: bytes) -> typing.Tuple[int, int]:
        """ returns : tuple (decoded int, total length) """
        if raw[0] != cls._TYPE_INTEGER:
            raise ValueError(f"Not an integer : {raw}")
        data_len, pos_data = cls.__get_asn1_ber_len(raw)
        val = 0
        for b in raw[pos_data:pos_data + data_len]:  # pylint: disable=invalid-name
            val = val * 256 + b
        return val, (pos_data + data_len)

    @classmethod
    def transform_value_to_int(cls, raw: bytes) -> int:  # noqa: E501 pylint: disable=missing-function-docstring
        data, total_len = cls._transform_value_to_int_no_len_check(raw)
        if total_len != len(raw):
            raise ValueError(
                f"wrong extension length : {raw} , found {total_len}, expected {len(raw)}"
            )
        return data

    @classmethod
    def _transform_value_to_sequence_no_len_check(cls, raw: bytes) -> typing.Tuple[list, int]:
        """ returns : tuple (decoded list, total length) """
        if raw[0] not in (cls._TYPE_SEQUENCE, cls._TYPE_SEQUENCE_OF):
            raise ValueError(f"Not a sequence : {raw}")
        data_len, pos_data = cls.__get_asn1_ber_len(raw)

        indefinite_len = bool(data_len == 0 and raw[1] == 0x80)
        decoded_list = []
        current_pos = pos_data
        while True:
            if indefinite_len:
                # Indefinite length : the end is indicated by the two bytes 00 00
                if raw[current_pos] == 0 and raw[current_pos + 1] == 0:
                    current_pos += 2
                    if current_pos != len(raw):
                        raise ValueError(
                            f"wrong extension length : {raw} , "
                            f"indefinite len ending at position {data_len}, expected {len(raw)}"
                        )
                    break
            else:
                if current_pos == pos_data + data_len:
                    break

                if current_pos > pos_data + data_len:
                    raise IndexError(
                        f"Error parsing data : current_pos = {current_pos} / "
                        f"pos_data = {pos_data} / data_len = {data_len} / raw = {raw}"
                    )

            if raw[current_pos] == cls._TYPE_INTEGER:
                tmp, tmp_len = cls._transform_value_to_int_no_len_check(raw[current_pos:])
                decoded_list.append(tmp)
                current_pos += tmp_len
            elif raw[current_pos] == cls._TYPE_OCTET_STR:
                tmp, tmp_len = cls._transform_value_to_str_no_len_check(raw[current_pos:])
                decoded_list.append(tmp)
                current_pos += tmp_len
            elif raw[current_pos] in (cls._TYPE_SEQUENCE, cls._TYPE_SEQUENCE_OF):
                tmp, tmp_len = cls._transform_value_to_sequence_no_len_check(raw[current_pos:])
                decoded_list.append(tmp)
                current_pos += tmp_len
            else:
                raise NotImplementedError(
                    f"Unknown type found : 0x{raw[current_pos]:02x} "
                    f"at position {current_pos} in raw = {raw}"
                )
        return decoded_list, current_pos

    @classmethod
    def transform_value_to_sequence(cls, raw: bytes) -> list:  # noqa: E501 pylint: disable=missing-function-docstring
        data, total_len = cls._transform_value_to_sequence_no_len_check(raw)
        if total_len != len(raw):
            raise ValueError(
                f"wrong extension length : {raw} , found {total_len}, expected {len(raw)}"
            )
        return data


class Extension:  # pylint: disable=missing-class-docstring

    def __init__(self, cert_ext: cryptography.x509.extensions.Extension):
        self._cert_ext = cert_ext

    @property
    def critical(self) -> bool:  # pylint: disable=missing-function-docstring
        return self._cert_ext.critical

    @property
    def oid(self) -> str:  # pylint: disable=missing-function-docstring
        return self._cert_ext.oid.dotted_string

    @property
    def value(self):
        """
        raw ASN1 value (bytes) : self.value.value
        """
        return self._cert_ext.value.value

    @property
    def raw(self):
        """
        Examples :
        OID as string : self.raw.oid.dotted_string
        raw ASN1 value (bytes) : self.raw.value.value
        """
        return self._cert_ext

    @property
    def value_as_str(self) -> str:  # pylint: disable=missing-function-docstring
        return Asn1BerDecoder.transform_value_to_str(self.value)

    @property
    def value_as_int(self) -> int:  # pylint: disable=missing-function-docstring
        return Asn1BerDecoder.transform_value_to_int(self.value)

    @property
    def value_as_sequence(self) -> list:  # pylint: disable=missing-function-docstring
        return Asn1BerDecoder.transform_value_to_sequence(self.value)

    def __str__(self):
        return str(self._cert_ext)

    def __repr__(self):
        return repr(self._cert_ext)


class ExtName(enum.Enum):  # pylint: disable=missing-class-docstring

    # https://confluence.protontech.ch/display/VPN/Agent+features+directory+and+format

    _TWO_FACTORS = "0.0.0"
    USER_TIER = "0.0.1"
    GROUPS = "0.0.2"
    PLATFORM = "0.0.3"
    NETSHIELD = "0.1.0"
    PORT_FW = "0.1.3"
    JAIL = "0.1.5"
    SPLIT_TCP = "0.1.6"
    RANDOM_NAT = "0.1.7"
    BOUNCING = "0.1.8"
    SAFE_MODE = "0.1.9"


class Certificate:  # pylint: disable=missing-class-docstring

    PROTONVPN_OID_STR = '1.3.6.1.4.1.56809.1'
    PROTONVPN_OID_ARRAY = PROTONVPN_OID_STR.split(".")

    def __init__(self, cert_pem: typing.Union[bytes, str] = None, cert_der: bytes = None):

        cert_input = [(cert_pem, "PEM"), (cert_der, "DER")]
        cert_input = [(x, x_type) for x, x_type in cert_input if x is not None]

        if len(cert_input) > 1:
            raise ValueError(
                "Not possible to provide multiple cert format. "
                f"Provided formats = {'/'.join([x_type for _, x_type in cert_input])}"
            )

        backend_x509 = None

        # cryptography.sys.version_info not available in 2.6
        crypto_major, crypto_minor = cryptography.__version__.split(".")[:2]

        if (
            int(crypto_major) < 3
            or int(crypto_major) == 3 and int(crypto_minor) < 1
        ):
            # backend is required if library < 3.1
            backend_x509 = cryptography.hazmat.backends.default_backend()

        if cert_pem is not None:
            if isinstance(cert_pem, str):
                cert_pem = cert_pem.encode("ascii")
            self._cert = cryptography.x509.load_pem_x509_certificate(
                data=cert_pem, backend=backend_x509
            )
        elif cert_der is not None:
            self._cert = cryptography.x509.load_der_x509_certificate(
                data=cert_der, backend=backend_x509
            )
        else:
            raise ValueError("Not provided any cert format")

    @property
    def raw(self):  # pylint: disable=missing-function-docstring
        return self._cert

    @property
    def public_key(self) -> bytes:  # pylint: disable=missing-function-docstring
        return self._cert.public_key().public_bytes(encoding=Encoding.Raw, format=PublicFormat.Raw)

    @property
    def proton_fingerprint(self) -> str:  # pylint: disable=missing-function-docstring
        ed25519_pk = self.public_key
        x25519_pk = nacl.bindings.crypto_sign_ed25519_pk_to_curve25519(ed25519_pk)
        return self.get_proton_fingerprint_from_x25519_pk(x25519_pk)

    @property
    def has_valid_date(self) -> bool:  # pylint: disable=missing-function-docstring
        return self.validity_period >= 0

    @property
    def validity_period(self) -> float:
        """ remaining time the certificate is valid,
            in seconds. < 0 : certificate is not valid anymore.
        """
        now_timestamp = datetime.datetime.now(datetime.timezone.utc).timestamp()
        return self.validity_date.timestamp() - now_timestamp

    @property
    def validity_date(self) -> datetime.datetime:  # pylint: disable=missing-function-docstring
        # cryptography >= v42.0.0 added `not_valid_after_utc` and deprecated `not_valid_after`.
        if hasattr(self._cert, "not_valid_after_utc"):
            return self._cert.not_valid_after_utc

        # Because `not_valid_after` returns a naive utc
        # datetime object (without time zone info), we add it manually.
        return self._cert.not_valid_after.replace(
            tzinfo=datetime.timezone.utc
        )

    @property
    def issued_date(self) -> datetime.datetime:  # pylint: disable=missing-function-docstring
        # cryptography >= v42.0.0 added `not_valid_before_utc` and deprecated `not_valid_before`.
        if hasattr(self._cert, "not_valid_before_utc"):
            return self._cert.not_valid_before_utc

        # Because `not_valid_before` returns a naive utc
        # datetime object (without time zone info), we add it manually.
        return self._cert.not_valid_before.replace(tzinfo=datetime.timezone.utc)

    @property
    def duration(self) -> datetime.timedelta:
        """ certification duration """
        return self.validity_date - self.issued_date

    @classmethod
    def get_proton_fingerprint_from_x25519_pk(cls, x25519_pk: bytes) -> str:  # noqa: E501 pylint: disable=missing-function-docstring
        return base64.b64encode(hashlib.sha512(x25519_pk).digest()).decode("ascii")

    def get_as_der(self) -> bytes:  # pylint: disable=missing-function-docstring
        return self._cert.public_bytes(Encoding.DER)

    def get_as_pem(self) -> str:  # pylint: disable=missing-function-docstring
        return self._cert.public_bytes(Encoding.PEM).decode("ascii")

    @property
    def proton_extensions(self) -> typing.Dict[ExtName, Extension]:  # noqa: E501 pylint: disable=missing-function-docstring
        extensions = {}
        for ext in self._cert.extensions:
            oid_array = ext.oid.dotted_string.split(".")
            if oid_array[:len(self.PROTONVPN_OID_ARRAY)] == self.PROTONVPN_OID_ARRAY:
                try:
                    ext_name = ".".join(oid_array[len(self.PROTONVPN_OID_ARRAY):])
                    ext_name = ExtName(ext_name)
                except ValueError:
                    continue
                extensions[ext_name] = Extension(ext)
        return extensions



"""
This modules contains the classes that communicate with NetworkManager.


Copyright (c) 2023 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
from ipaddress import ip_network
import asyncio
import concurrent.futures

from proton.vpn import logging
from proton.vpn.backend.networkmanager.killswitch.default.nmclient import NMClient
from proton.vpn.backend.networkmanager.killswitch.default.killswitch_connection import (
    KillSwitchConnection, KillSwitchGeneralConfig, KillSwitchIPConfig
)

logger = logging.getLogger(__name__)

LOCAL_AGENT_SERVER_ADDR = "10.2.0.1"  # 10.2.0.1:65432


def _get_connection_id(prefix: str, permanent: bool, ipv6: bool = False, routed: bool = False):
    if ipv6:
        return f"{prefix}-killswitch-ipv6{'-perm' if permanent else ''}"

    return f"{prefix}{'-routed' if routed else ''}-killswitch{'-perm' if permanent else ''}"


def _get_interface_name(permanent: bool, ipv6: bool = False, routed: bool = False):
    if ipv6:
        return f"ipv6leakintrf{'1' if permanent else '0'}"

    return f"{'pvpnrouteintrf' if routed else 'pvpnksintrf'}{'1' if permanent else '0'}"


async def _wrap_future(future: concurrent.futures.Future, timeout=5):
    """Wraps a concurrent.future.Future object in an asyncio.Future object."""
    return await asyncio.wait_for(
        asyncio.wrap_future(future, loop=asyncio.get_running_loop()),
        timeout=timeout
    )


def build_routes_list(server_ip: str):
    """
    Builds a list of routes to block all traffic except the server IP and
    the local agent server IP.
    """
    local_agent_network = ip_network(LOCAL_AGENT_SERVER_ADDR)
    most_networks = ip_network('0.0.0.0/0').address_exclude(
        ip_network(server_ip))

    for network_a in most_networks:
        if local_agent_network.overlaps(network_a):
            for network_b in network_a.address_exclude(local_agent_network):
                yield network_b
        else:
            yield network_a


class KillSwitchConnectionHandler:
    """Kill switch connection management."""

    def __init__(self, nm_client: NMClient = None, connection_prefix: str = None):
        self._nm_client = nm_client
        self._connection_prefix = connection_prefix or "pvpn"
        self._ipv6_ks_settings = KillSwitchIPConfig(
            addresses=["fdeb:446c:912d:08da::/64"],
            dns=["::1"],
            dns_priority=-1400,
            gateway="fdeb:446c:912d:08da::1",
            ignore_auto_dns=True,
            route_metric=95
        )

    @staticmethod
    def _get_ipv4_ks_settings(server_ip: str = None):
        if server_ip:
            # accept/block all routes except the server IP route.
            routes = list(build_routes_list(server_ip))
            gateway = None
        else:
            routes = []  # accept/block all routes.
            gateway = "100.85.0.1"

        return KillSwitchIPConfig(
            addresses=["100.85.0.1/24"],
            dns=["0.0.0.0"],  # nosec hardcoded_bind_all_interfaces
            dns_priority=-1400,
            gateway=gateway,
            ignore_auto_dns=True,
            route_metric=98,
            routes=routes
        )

    @property
    def nm_client(self):
        """Returns the NetworkManager client."""
        if self._nm_client is None:
            self._nm_client = NMClient()

        return self._nm_client

    @property
    def is_network_manager_running(self) -> bool:
        """Returns if the Network Manager daemon is running or not."""
        return self.nm_client.get_nm_running()

    @property
    def is_connectivity_check_enabled(self) -> bool:
        """Returns if connectivity_check property is enabled or not."""
        return self.nm_client.connectivity_check_get_enabled()

    async def add_full_killswitch_connection(self, permanent: bool):
        """Adds full kill switch connection to Network Manager. This connection blocks all
        outgoing traffic when not connected to VPN, with the exception of torrent client which will
        require to be bonded to the VPN interface.."""
        await self._ensure_connectivity_check_is_disabled()

        connection_id = _get_connection_id(self._connection_prefix, permanent)
        connection = self.nm_client.get_active_connection(
            conn_id=connection_id
        )

        if connection:
            logger.debug("Kill switch was already present.")
            return

        interface_name = _get_interface_name(permanent)
        general_config = KillSwitchGeneralConfig(
            human_readable_id=connection_id,
            interface_name=interface_name
        )

        kill_switch = KillSwitchConnection(
            general_config,
            ipv4_settings=self._get_ipv4_ks_settings(),
            ipv6_settings=self._ipv6_ks_settings,
        )
        await _wrap_future(
            self.nm_client.add_connection_async(kill_switch.connection, save_to_disk=permanent)
        )

        logger.debug(f"{'Permanent' if permanent else 'Non-permanent'} kill switch added.")
        await self._remove_connection(
            connection_id=_get_connection_id(self._connection_prefix, permanent=not permanent)
        )
        logger.debug(f"{'Non-permanent' if permanent else 'Permanent'} kill switch removed.")

    async def add_routed_killswitch_connection(self, server_ip: str, permanent: bool):
        """Add routed kill switch connection to Network Manager.

        This connection has a "hole punched in it", to allow only the server IP to
        access the outside world while blocking all other outgoing traffic. This is only
        temporary though as it will be removed once we establish a VPN connection and will
        get replaced by the full kill switch connection.
        """
        await self._ensure_connectivity_check_is_disabled()

        general_config = KillSwitchGeneralConfig(
            human_readable_id=_get_connection_id(self._connection_prefix, permanent, routed=True),
            interface_name=_get_interface_name(permanent, routed=True)
        )
        kill_switch = KillSwitchConnection(
            general_config,
            ipv4_settings=self._get_ipv4_ks_settings(server_ip),
            ipv6_settings=self._ipv6_ks_settings,
        )
        await _wrap_future(
            self.nm_client.add_connection_async(kill_switch.connection, save_to_disk=permanent)
        )
        logger.debug("Routed kill switch added.")

    async def add_ipv6_leak_protection(self):
        """Adds IPv6 kill switch to NetworkManager. This connection is mainly
        to prevent IPv6 leaks while using IPv4."""
        await self._ensure_connectivity_check_is_disabled()

        connection_id = _get_connection_id(
            self._connection_prefix, permanent=False, ipv6=True
        )
        connection = self.nm_client.get_active_connection(
            conn_id=connection_id)

        if connection:
            logger.debug("IPv6 leak protection already present.")
            return

        interface_name = _get_interface_name(permanent=False, ipv6=True)
        general_config = KillSwitchGeneralConfig(
            human_readable_id=connection_id,
            interface_name=interface_name
        )

        kill_switch = KillSwitchConnection(
            general_config,
            ipv4_settings=None,
            ipv6_settings=self._ipv6_ks_settings,
        )

        await _wrap_future(
            self.nm_client.add_connection_async(kill_switch.connection, save_to_disk=False)
        )
        logger.debug("IPv6 leak protection added.")

    async def remove_full_killswitch_connection(self):
        """Removes full kill switch connection."""
        logger.debug("Removing full kill switch...")
        await self._remove_connection(
            _get_connection_id(self._connection_prefix, permanent=True)
        )
        await self._remove_connection(
            _get_connection_id(self._connection_prefix, permanent=False)
        )
        logger.debug("Full kill switch removed.")

    async def remove_routed_killswitch_connection(self):
        """Removes routed kill switch connection."""
        logger.debug("Removing routed kill switch...")
        await self._remove_connection(
            _get_connection_id(self._connection_prefix, permanent=True, routed=True)
        )
        await self._remove_connection(
            _get_connection_id(self._connection_prefix, permanent=False, routed=True)
        )
        logger.debug("Routed kill switch removed.")

    async def remove_ipv6_leak_protection(self):
        """Removes IPv6 kill switch connection."""
        logger.debug("Removing IPv6 leak protection...")
        await self._remove_connection(
            _get_connection_id(self._connection_prefix, permanent=False, ipv6=True)
        )
        logger.debug("IP6 leak protection removed.")

    async def _remove_connection(self, connection_id: str):
        connection = self.nm_client.get_connection(
            conn_id=connection_id)

        logger.debug(f"Attempting to remove {connection_id}: {connection}")

        if not connection:
            logger.debug(f"There was no {connection_id} to remove")
            return

        await _wrap_future(self.nm_client.remove_connection_async(connection))

    async def _ensure_connectivity_check_is_disabled(self):
        if self.is_connectivity_check_enabled:  # noqa: E501 # pylint: disable=line-too-long # nosemgrep: python.lang.maintainability.is-function-without-parentheses.is-function-without-parentheses
            await _wrap_future(self.nm_client.disable_connectivity_check())
            logger.info("Network connectivity check was disabled.")



"""
Copyright (c) 2023 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""

from proton.vpn.session.dataclasses.bug_report import BugReportForm
from proton.vpn.session.dataclasses.certificate import VPNCertificate
from proton.vpn.session.dataclasses.credentials import (
    VPNUserPassCredentials, VPNCredentials
)
from proton.vpn.session.dataclasses.location import VPNLocation
from proton.vpn.session.dataclasses.login_result import LoginResult
from proton.vpn.session.dataclasses.sessions import APIVPNSession, VPNSessions
from proton.vpn.session.dataclasses.settings import VPNInfo, VPNSettings

__all__ = [
    "BugReportForm",
    "VPNCertificate",
    "VPNUserPassCredentials", "VPNCredentials",
    "VPNLocation",
    "LoginResult",
    "APIVPNSession", "VPNSessions",
    "VPNInfo", "VPNSettings"
]



"""
Certain VPN data like the server list and the client configuration needs to
refreshed periodically to keep it up to date.

This module defines the required services to do so.


Copyright (c) 2023 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
from datetime import timedelta
from typing import Callable, Optional

from proton.vpn import logging
from proton.vpn.core.refresher.certificate_refresher import CertificateRefresher
from proton.vpn.core.refresher.client_config_refresher import ClientConfigRefresher
from proton.vpn.core.refresher.feature_flags_refresher import FeatureFlagsRefresher
from proton.vpn.core.refresher.scheduler import Scheduler
from proton.vpn.core.refresher.server_list_refresher import ServerListRefresher
from proton.vpn.core.session_holder import SessionHolder
from proton.vpn.session.client_config import ClientConfig
from proton.vpn.session import FeatureFlags
from proton.vpn.session.servers.logicals import ServerList

logger = logging.getLogger(__name__)


class VPNDataRefresher:  # pylint: disable=too-many-instance-attributes
    """
    Service in charge of:
        - retrieving the required VPN data from Proton's REST API
          to be able to establish VPN connection,
        - keeping it up to date and
        - notifying subscribers when VPN data has been updated.
    """
    def __init__(  # pylint: disable=too-many-arguments
        self,
        session_holder: SessionHolder,
        scheduler: Scheduler,
        client_config_refresher: ClientConfigRefresher = None,
        server_list_refresher: ServerListRefresher = None,
        certificate_refresher: CertificateRefresher = None,
        feature_flags_refresher: FeatureFlagsRefresher = None,
    ):
        self._session_holder = session_holder
        self._scheduler = scheduler
        self._client_config_refresher = client_config_refresher or ClientConfigRefresher(
            session_holder
        )
        self._server_list_refresher = server_list_refresher or ServerListRefresher(
            session_holder
        )
        self._certificate_refresher = certificate_refresher or CertificateRefresher(
            session_holder
        )
        self._feature_flags_refresher = feature_flags_refresher or FeatureFlagsRefresher(
            session_holder
        )
        self._client_config_refresh_task_id = None
        self._server_list_refresher_task_id = None
        self._certificate_refresher_task_id = None
        self._feature_flags_refresher_task_id = None

    def set_error_callback(self, error_callback: Callable[[Exception], None] = None):
        """Sets the error callback to be called when an error occurs while executing a task."""
        self._scheduler.set_error_callback(error_callback)

    def unset_error_callback(self):
        """Unsets the error callback."""
        self._scheduler.unset_error_callback()

    @property
    def _session(self):
        return self._session_holder.session

    def set_server_list_updated_callback(self, callback: Optional[Callable]):
        """Sets the callback to be called whenever the server list is updated."""
        self._server_list_refresher.server_list_updated_callback = callback

    def set_server_loads_updated_callback(self, callback: Optional[Callable]):
        """Sets the callback to be called whenever the server loads are updated."""
        self._server_list_refresher.server_loads_updated_callback = callback

    def set_certificate_updated_callback(self, callback: Optional[Callable]):
        """Sets the callback to be called whenever the certificate is updated."""
        self._certificate_refresher.certificate_updated_callback = callback

    async def get_up_to_date_server_list(self) -> ServerList:
        """
        Returns the list of available VPN servers, after updating if expired.
        """
        await self._refresh_vpn_session_if_necessary()
        if not self._scheduler.is_started:
            # only update if scheduler isn't tasked with doing so
            await self._server_list_refresher.update_if_necessary()

        return self._session.server_list

    async def get_up_to_date_client_config(self) -> ClientConfig:
        """
        Returns the VPN client configuration, after updating if expired.
        """
        await self._refresh_vpn_session_if_necessary()
        if not self._scheduler.is_started:
            # only update if scheduler isn't tasked with doing so
            await self._client_config_refresher.update_if_necessary()

        return self._session.client_config

    async def get_up_to_date_feature_flags(self) -> FeatureFlags:
        """
        Returns the feature flags list, after updating if expired.
        """
        await self._refresh_vpn_session_if_necessary()
        if not self._scheduler.is_started:
            # only update if scheduler isn't tasked with doing so
            await self._feature_flags_refresher.update_if_necessary()

        return self._session.feature_flags

    async def update_certificate_if_necessary(self):
        """
        Updates the API certificate if it has expired, or will soon do so.
        """
        await self._refresh_vpn_session_if_necessary()
        if not self._scheduler.is_started:
            # only update if scheduler isn't tasked with doing so
            await self._certificate_refresher.update_if_necessary()

    @property
    def server_list(self) -> ServerList:
        """
        Returns the list of available VPN servers.
        """
        return self._session.server_list

    @property
    def client_config(self) -> ClientConfig:
        """Returns the VPN client configuration."""
        return self._session.client_config

    @property
    def feature_flags(self) -> FeatureFlags:
        """Returns VPN features."""
        return self._session.feature_flags

    def force_refresh_certificate(self):
        """Force refresh certificate on demand."""
        logger.info("Force refresh certificate.")
        self._scheduler.cancel_task(self._certificate_refresher_task_id)
        self._certificate_refresher_task_id = self._scheduler.run_soon(
            self._certificate_refresher.refresh
        )

    @property
    def is_vpn_data_ready(self) -> bool:
        """Returns whether the necessary data from API has already been retrieved or not."""
        return self._session.loaded

    async def enable(self):
        """Start retrieving data periodically from Proton's REST API."""
        await self._refresh_vpn_session_if_necessary()
        self._enable()

    async def disable(self):
        """Stops retrieving data periodically from Proton's REST API."""
        self._scheduler.cancel_task(self._client_config_refresh_task_id)
        self._client_config_refresh_task_id = None

        self._scheduler.cancel_task(self._server_list_refresher_task_id)
        self._server_list_refresher_task_id = None

        self._scheduler.cancel_task(self._certificate_refresher_task_id)
        self._certificate_refresher_task_id = None

        self._scheduler.cancel_task(self._feature_flags_refresher_task_id)
        self._feature_flags_refresher_task_id = None

        await self._scheduler.stop()
        logger.info(
            "VPN data refresher service disabled.",
            category="app", subcategory="vpn_data_refresher", event="disable"
        )

    def _enable(self):
        logger.info(
            "VPN data refresher service enabled.",
            category="app", subcategory="vpn_data_refresher", event="enable"
        )
        self._client_config_refresh_task_id = self._scheduler.run_after(
            self._client_config_refresher.initial_refresh_delay,
            self._client_config_refresher.refresh
        )
        logger.info(
            f"Next client config refresh scheduled in "
            f"{timedelta(seconds=self._client_config_refresher.initial_refresh_delay)}"
        )

        self._server_list_refresher_task_id = self._scheduler.run_after(
            self._server_list_refresher.initial_refresh_delay,
            self._server_list_refresher.refresh
        )
        logger.info(
            f"Next server list refresh scheduled in "
            f"{timedelta(seconds=self._server_list_refresher.initial_refresh_delay)}"
        )

        self._certificate_refresher_task_id = self._scheduler.run_after(
            self._certificate_refresher.initial_refresh_delay,
            self._certificate_refresher.refresh
        )
        logger.info(
            f"Next certificate refresh scheduled in "
            f"{timedelta(seconds=self._certificate_refresher.initial_refresh_delay)}"
        )

        self._feature_flags_refresher_task_id = self._scheduler.run_after(
            self._feature_flags_refresher.initial_refresh_delay,
            self._feature_flags_refresher.refresh
        )
        logger.info(
            f"Next feature flags refresh scheduled in "
            f"{timedelta(seconds=self._feature_flags_refresher.initial_refresh_delay)}"
        )

        self._scheduler.start()

    async def _refresh_vpn_session_if_necessary(self):
        if not self._session.loaded:
            # The VPN session is normally loaded straight after the user logs in. However,
            # it could happen that it's not loaded in any of the following scenarios:
            # a) After a successful authentication, the HTTP requests to retrieve
            #    the required VPN session data failed, so it was never persisted.
            # b) The persisted VPN session does not have the expected format.
            #    This can happen if we introduce a breaking change or if the persisted
            #    data is messed up because the user changes it, or it gets corrupted.
            logger.warning("Reloading VPN session...")
            await self._session.fetch_session_data()



"""
Copyright (c) 2023 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
from __future__ import annotations
from typing import Iterable, List, Optional, Protocol, TYPE_CHECKING
from dataclasses import dataclass
import itertools

from proton.vpn.session.servers.country_codes import get_country_name_by_code
from proton.vpn.session.servers.types import ServerFeatureEnum

if TYPE_CHECKING:
    from proton.vpn.session.servers.logicals import LogicalServer


class ServerGroup(Protocol):  # pylint: disable=too-few-public-methods
    """A group of servers."""
    name: str
    servers: list[LogicalServer]
    free: bool  # Whether the group has servers available to the free tier or not
    features: set[ServerFeatureEnum]
    under_maintenance: bool
    smart_routing: bool
    free_servers: list[LogicalServer]
    paid_servers: list[LogicalServer]


SECURE_CORE_GROUP_NAME = "Via Secure Core"


class SecureCoreGroup(ServerGroup):
    """A group of secure core servers. Secure Core is paid only."""

    def __init__(self, servers: list[LogicalServer]):
        """
        :param servers: List of logical servers in this secure core group.
        """
        self._servers = list(servers)
        self._analysis = ServerAnalysis.analyze_servers(self._servers)

    @property
    def name(self) -> str:
        """Returns the secure core group name."""
        return SECURE_CORE_GROUP_NAME

    @property
    def servers(self) -> List[LogicalServer]:
        """Returns the list of logical servers in this secure core group."""
        return self._servers

    @property
    def free(self) -> bool:
        """Returns whether the group has servers available to the free tier or not."""
        return self._analysis.free

    @property
    def features(self) -> set[ServerFeatureEnum]:
        """Returns the set of features available in this group."""
        return {ServerFeatureEnum.SECURE_CORE}

    @property
    def under_maintenance(self) -> bool:
        """Returns whether all servers in this group are under maintenance."""
        return self._analysis.under_maintenance

    @property
    def smart_routing(self) -> bool:
        """Returns whether all servers in this group use smart routing."""
        return self._analysis.smart_routing

    @property
    def free_servers(self) -> Iterable[LogicalServer]:
        """Returns the free servers only."""
        return filter(lambda server: server.free, self._servers)

    @property
    def paid_servers(self) -> Iterable[LogicalServer]:
        """Returns the paid servers only."""
        return filter(lambda server: not server.free, self._servers)


class City(ServerGroup):
    """A city that belongs to a country."""

    def __init__(self, name: str, servers: list[LogicalServer]):
        self._name = name
        self._servers = servers
        self._analysis = ServerAnalysis.analyze_servers(servers)

    @property
    def name(self) -> str:
        """Returns the city name."""
        return self._name

    @property
    def servers(self) -> list[LogicalServer]:
        """Returns the list of logical servers in this city."""
        return self._servers

    @property
    def free(self) -> bool:
        """Returns whether the city has servers available to the free tier or not."""
        return self._analysis.free

    @property
    def features(self) -> set[ServerFeatureEnum]:
        """Returns the set of features available in this city."""
        return self._analysis.features

    @property
    def under_maintenance(self) -> bool:
        """Returns whether all servers in this city are under maintenance."""
        return self._analysis.under_maintenance

    @property
    def smart_routing(self) -> bool:
        """Returns whether all servers in this city use smart routing."""
        return self._analysis.smart_routing

    @property
    def free_servers(self) -> Iterable[LogicalServer]:
        """Returns the free servers only."""
        return filter(lambda server: server.free, self._servers)

    @property
    def paid_servers(self) -> Iterable[LogicalServer]:
        """Returns the paid servers only."""
        return filter(lambda server: not server.free, self._servers)


class Country(ServerGroup):  # pylint: disable=too-many-instance-attributes
    """Group of servers belonging to a country."""

    def __init__(self, code: str, servers: List[LogicalServer]):
        """
        :param code: The country code (ISO 3166-1 alpha-2).
        :param servers: List of logical servers in this country sorted by city name.
        """
        self._code = code
        self._servers = servers
        self._analysis = ServerAnalysis.analyze_servers(servers)
        self._cities = self._build_cities(servers)
        self._secure_core_group = self._build_secure_core_group(servers)

    @property
    def code(self) -> str:
        """Returns the country code (ISO 3166-1 alpha-2)."""
        return self._code

    @property
    def name(self):
        """Returns the full country name."""
        return get_country_name_by_code(self.code)

    @property
    def servers(self) -> List[LogicalServer]:
        """Returns the list of logical servers in this country."""
        return self._servers

    @property
    def cities(self) -> list[City]:
        """Returns the list of cities available in the country."""
        return self._cities

    @property
    def free(self) -> bool:
        """Returns whether the country has servers available to the free tier or not."""
        return self._analysis.free

    @property
    def features(self) -> set[ServerFeatureEnum]:
        """Returns the set of features available in this country."""
        return self._analysis.features

    @property
    def under_maintenance(self) -> bool:
        """Returns whether all servers in this country are under maintenance."""
        return self._analysis.under_maintenance

    @property
    def smart_routing(self) -> bool:
        """Returns whether all servers in this country use smart routing."""
        return self._analysis.smart_routing

    @property
    def free_cities(self) -> Iterable[City]:
        """Returns the free cities only."""
        return filter(lambda city: city.free, self.cities)

    @property
    def paid_cities(self) -> Iterable[City]:
        """Returns the paid cities only."""
        return filter(lambda city: not city.free, self.cities)

    @property
    def secure_core_group(self) -> Optional[SecureCoreGroup]:
        """Returns the secure core servers group."""
        return self._secure_core_group

    def _build_cities(self, servers: List[LogicalServer]) -> list[City]:
        """Returns the list of cities available in the country."""
        non_secure_core_servers = filter(
            lambda servers: ServerFeatureEnum.SECURE_CORE not in servers.features, servers
        )
        return [
            City(city_name, list(city_servers))
            for city_name, city_servers in itertools.groupby(
                non_secure_core_servers,
                key=lambda server: server.city
            )
        ]

    def _build_secure_core_group(self, servers: List[LogicalServer]) -> SecureCoreGroup:
        secure_core_servers = list(filter(
            lambda server: ServerFeatureEnum.SECURE_CORE in server.features, self.servers
        ))

        if not secure_core_servers:
            return None

        return SecureCoreGroup(secure_core_servers)


@dataclass
class ServerAnalysis:
    """Contains a summary of the state all the servers in a given location."""
    smart_routing: bool
    under_maintenance: bool
    free: bool
    features: set[ServerFeatureEnum]

    @staticmethod
    def analyze_servers(ordered_servers: List[LogicalServer]) -> ServerAnalysis:
        """
        Iterates over the ordered list of servers and extracts information
        to be displayed on the server list row.
        """
        free_location = None

        features = set()

        # The country is set under maintenance until the opposite is proven.
        under_maintenance = True

        # Smart routing is assumed to be used until the opposite is proven.
        smart_routing_location = True

        for server in ordered_servers:
            features.update(server.features)

            free_location = free_location or server.tier == 0

            # The country is under maintenance if (1) that was the case up until now and
            # (2) the current server is also under maintenance (i.e. is not enabled).
            under_maintenance = (under_maintenance and not server.enabled)

            # A country is flagged as a "Smart routing" location if *all* servers are
            # actually physically located in a neighboring country.
            smart_routing_location = \
                (smart_routing_location and server.smart_routing)

        return ServerAnalysis(
            smart_routing_location,
            under_maintenance,
            free_location,
            features
        )



"""
Connection persistence.

Connection parameters are persisted to disk so that they can be loaded after a crash.


Copyright (c) 2023 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
from __future__ import annotations
import json
import os
from dataclasses import dataclass
from json import JSONDecodeError
from typing import Optional

from proton.utils.environment import VPNExecutionEnvironment
from proton.vpn import logging
from proton.vpn.connection.interfaces import VPNServer

logger = logging.getLogger(__name__)


@dataclass
class ConnectionParameters:
    """Connection parameters to be persisted to disk."""
    connection_id: str
    backend: str
    protocol: str
    server: VPNServer

    @classmethod
    def from_dict(cls, data: dict) -> ConnectionParameters:
        """Creates a ConnectionParameters instance from a dictionary."""
        return cls(
            connection_id=data["connection_id"],
            backend=data["backend"],
            protocol=data["protocol"],
            server=VPNServer.from_dict(data["server"])
        )

    def to_dict(self) -> ConnectionParameters:
        """Creates a dictionary from a ConnectionParameters instance."""
        return {
            "connection_id": self.connection_id,
            "backend": self.backend,
            "protocol": self.protocol,
            "server": self.server.to_dict()
        }


class ConnectionPersistence:
    """Saves/loads connection parameters to/from disk."""
    FILENAME = "connection_persistence.json"

    def __init__(self, persistence_directory: str = None):
        self._directory = persistence_directory

    @property
    def _connection_file_path(self):
        if not self._directory:
            self._directory = os.path.join(
                VPNExecutionEnvironment().path_cache, "connection"
            )
            os.makedirs(self._directory, mode=0o700, exist_ok=True)

        return os.path.join(self._directory, self.FILENAME)

    def load(self) -> Optional[ConnectionParameters]:
        """Returns the connection parameters loaded from disk, or None if
        no connection parameters were persisted yet."""
        if not os.path.isfile(self._connection_file_path):
            return None

        with open(self._connection_file_path, encoding="utf-8") as file:
            try:
                file_content = json.load(file)
                return ConnectionParameters.from_dict(file_content)
            except (JSONDecodeError, KeyError, UnicodeDecodeError):
                logger.warning(
                    "Unexpected error parsing connection persistence file: "
                    f"{self._connection_file_path}",
                    category="CONN", subcategory="PERSISTENCE", event="LOAD",
                    exc_info=True
                )
                return None

    def save(self, connection_parameters: ConnectionParameters):
        """Saves connection parameters to disk."""
        with open(self._connection_file_path, "w", encoding="utf-8") as file:
            json.dump(connection_parameters.to_dict(), file)

    def remove(self):
        """Removes the connection persistence file, if it exists."""
        if os.path.isfile(self._connection_file_path):
            os.remove(self._connection_file_path)
        else:
            logger.warning(
                f"Connection persistence not found when trying "
                f"to remove it: {self._connection_file_path}",
                category="CONN", subcategory="PERSISTENCE", event="REMOVE"
            )



"""
VPN connector.


Copyright (c) 2023 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
from __future__ import annotations

import asyncio
from copy import deepcopy
import os
import threading
from typing import Optional, runtime_checkable, Protocol

from proton.loader import Loader
from proton.loader.loader import PluggableComponent

from proton.vpn.connection.persistence import ConnectionPersistence
from proton.vpn.core.refresher import VPNDataRefresher
from proton.vpn.core.session_holder import SessionHolder
from proton.vpn.core.settings import SettingsPersistence
from proton.vpn.core.settings.split_tunneling import SplitTunneling as SplitTunnelingSetting
from proton.vpn.killswitch.interface import KillSwitch

from proton.vpn import logging
from proton.vpn.connection import (
    events, states, VPNConnection, VPNServer, ProtocolPorts,
    VPNCredentials, Settings
)
from proton.vpn.connection.enum import KillSwitchSetting, ConnectionStateEnum
from proton.vpn.connection.publisher import Publisher
from proton.vpn.connection.states import StateContext
from proton.vpn.session.client_config import ClientConfig
from proton.vpn.session.dataclasses import VPNLocation
from proton.vpn.session.servers import LogicalServer, ServerFeatureEnum
from proton.vpn.core.usage import UsageReporting
from proton.vpn.connection.exceptions import FeatureSyntaxError, FeatureError
from proton.vpn.split_tunneling.interface import SplitTunneling
from proton.vpn.core.cache_handlers import PortForwardFileHandler

logger = logging.getLogger(__name__)


@runtime_checkable
class VPNStateSubscriber(Protocol):  # pylint: disable=too-few-public-methods
    """Subscriber to connection status updates."""

    def status_update(self, status: "BaseState"):  # noqa
        """This method is called by the publisher whenever a VPN connection status
        update occurs.
        :param status: new connection status.
        """


class VPNConnector:  # pylint: disable=too-many-instance-attributes
    """
    Allows connecting/disconnecting to/from Proton VPN servers, as well as querying
    information about the current VPN connection, or subscribing to its state
    updates.

    Multiple simultaneous VPN connections are not allowed. If a connection
    already exists when a new one is requested then the current one is brought
    down before starting the new one.
    """

    @classmethod
    async def get(  # pylint: disable=too-many-arguments
        cls,
        session_holder: SessionHolder,
        settings_persistence: SettingsPersistence,
        usage_reporting: UsageReporting,
        kill_switch: KillSwitch = None,
    ):
        """
        Builds a VPN connector instance and initializes it.
        """
        split_tunneling = await SplitTunneling.get(os.getuid())
        connector = VPNConnector(
            session_holder,
            settings_persistence,
            kill_switch=kill_switch,
            usage_reporting=usage_reporting,
            split_tunneling=split_tunneling
        )
        await connector.initialize_state()
        return connector

    def __init__(  # pylint: disable=too-many-arguments
            self,
            session_holder: SessionHolder,
            settings_persistence: SettingsPersistence,
            usage_reporting: UsageReporting,
            connection_persistence: Optional[ConnectionPersistence] = None,
            state: Optional[states.State] = None,
            kill_switch: Optional[KillSwitch] = None,
            split_tunneling: Optional[SplitTunneling] = None,
            publisher: Optional[Publisher] = None,
            port_forward_file_handler: PortForwardFileHandler = None,
    ):
        self._session_holder = session_holder
        self._settings_persistence = settings_persistence
        self._connection_persistence = connection_persistence or ConnectionPersistence()
        self._current_state = state
        self._kill_switch = kill_switch
        self._split_tunneling = split_tunneling
        self._publisher = publisher or Publisher()
        self._lock = asyncio.Lock()
        self._background_tasks = set()
        self._usage_reporting = usage_reporting
        self._port_forward_file_handler = port_forward_file_handler or PortForwardFileHandler()

        self._publisher.register(self._on_state_change_update_location)
        self._publisher.register(self._port_forward_file_handler.on_state_change_update_port)

    @property
    def is_split_tunneling_available(self) -> bool:
        """Returns if split tunneling is available or not."""
        return bool(self._split_tunneling)

    async def get_settings(self) -> Settings:
        """Returns the user's settings."""
        # Default to free user settings if the session is not loaded yet.
        user_tier = self._session_holder.user_tier or 0
        loop = asyncio.get_running_loop()
        settings = await loop.run_in_executor(
            None, self._settings_persistence.get,
            user_tier
        )

        return settings

    @property
    def credentials(self) -> Optional[VPNCredentials]:
        """Returns the user's credentials."""
        return self._session_holder.vpn_credentials

    def _set_ks_setting(self, ks_setting: KillSwitchSetting, protocol: str):
        StateContext.kill_switch_setting = ks_setting

        if isinstance(self.current_state, states.Disconnected):
            self._set_ks_impl(protocol)

    async def update_credentials(self):
        """
        Updates the credentials of the current connection.

        This is useful when the certificate used for the current connection
        has expired and a new one is needed.
        """
        if self.current_connection:
            logger.info("Updating credentials for current connection.")
            await self.current_connection.update_credentials(self.credentials)

    async def apply_settings(self, settings: Settings):
        """
        Sets the settings to be applied when establishing the next connection and
        applies them to the current connection whenever that's possible.
        """
        ks_setting = KillSwitchSetting(settings.killswitch)
        protocol = settings.protocol
        self._set_ks_setting(ks_setting, protocol)
        await self._apply_kill_switch_setting(ks_setting)

        if self.current_connection:
            await self.current_connection.update_settings(settings)

        st_setting = settings.features.split_tunneling
        self._set_split_tunneling_setting(st_setting)
        # nosemgrep: python.lang.maintainability.is-function-without-parentheses.is-function-without-parentheses  # pylint: disable=line-too-long  # noqa: E501
        if self.is_split_tunneling_available and self.is_connected:
            await self._apply_split_tunneling_settings(st_setting, ks_setting)

    async def _apply_kill_switch_setting(self, kill_switch_setting: KillSwitchSetting):
        """Enables/disables the kill switch depending on the setting value."""
        kill_switch = self._current_state.context.kill_switch

        if kill_switch_setting == KillSwitchSetting.PERMANENT:
            await kill_switch.enable(permanent=True)
            # Since full KS already prevents IPv6 leaks:
            await kill_switch.disable_ipv6_leak_protection()

        elif kill_switch_setting == KillSwitchSetting.ON:
            if isinstance(self._current_state, states.Disconnected):
                await kill_switch.disable()
                await kill_switch.disable_ipv6_leak_protection()
            else:
                await kill_switch.enable(permanent=False)
                # Since full KS already prevents IPv6 leaks:
                await kill_switch.disable_ipv6_leak_protection()

        elif kill_switch_setting == KillSwitchSetting.OFF:
            if isinstance(self._current_state, states.Disconnected):
                await kill_switch.disable()
                await kill_switch.disable_ipv6_leak_protection()
            else:
                await kill_switch.enable_ipv6_leak_protection()
                await kill_switch.disable()

        else:
            raise RuntimeError(f"Unexpected kill switch setting: {kill_switch_setting}")

    async def _apply_split_tunneling_settings(
            self, st_settings: SplitTunnelingSetting, ks_setting: KillSwitchSetting
    ):
        if ks_setting != KillSwitchSetting.OFF:
            logger.warning("Split tunneling is not compatible with the kill switch feature")
            return

        if not st_settings.enabled:
            await self._split_tunneling.clear_config()
        else:
            await self._split_tunneling.set_config(
                st_settings.get_config()
            )

    async def _get_current_connection(self) -> Optional[VPNConnection]:
        """
        :return: the current VPN connection or None if there isn't one.
        """
        loop = asyncio.get_running_loop()
        persisted_parameters = await loop.run_in_executor(None, self._connection_persistence.load)
        if not persisted_parameters:
            return None

        # I'm refraining of refactoring the whole thing but this way of loading
        # the protocol class is madness.
        backend_class = Loader.get("backend", persisted_parameters.backend)
        backend_name = backend_class.backend
        if persisted_parameters.backend != backend_name:
            return None

        all_protocols = Loader.get_all(backend_name)
        settings = await self.get_settings()
        for protocol in all_protocols:
            if protocol.cls.protocol == persisted_parameters.protocol:

                vpn_connection = protocol.cls(
                    server=persisted_parameters.server,
                    credentials=self.credentials,
                    settings=settings,
                    connection_id=persisted_parameters.connection_id
                )
                if not isinstance(vpn_connection.initial_state, states.Disconnected):
                    return vpn_connection

        return None

    async def _get_initial_state(self):
        """Determines the initial state of the state machine."""

        # It's possible that the user is not logged in but that there is
        # a persisted connection, in this case we need to ignore the persisted
        # connection and return the disconnected state.
        if self._session_holder.session.logged_in:
            current_connection = await self._get_current_connection()

            if current_connection:
                return current_connection.initial_state

        return states.Disconnected(
            StateContext(event=events.Initialized(events.EventContext(connection=None)))
        )

    async def initialize_state(self):
        """Initializes the state machine with the specified state."""
        state = await self._get_initial_state()

        settings = await self.get_settings()
        StateContext.kill_switch_setting = KillSwitchSetting(settings.killswitch)
        self._set_ks_impl(settings.protocol)

        self._set_split_tunneling_setting(settings.features.split_tunneling)
        self._set_split_tunneling_impl()

        connection = state.context.connection
        if connection:
            connection.register(self._on_connection_event)

        # Sets the initial state of the connector and triggers the tasks associated
        # to the state.
        await self._update_state(state)

        # Makes sure that the kill switch state is inline with the current
        # kill switch setting (e.g. if the KS setting is set to "permanent" then
        # the permanent KS should be enabled, if it was not the case yet).
        await self._apply_kill_switch_setting(StateContext.kill_switch_setting)

    @property
    def current_state(self) -> states.State:
        """Returns the state of the current VPN connection."""
        return self._current_state

    @property
    def current_connection(self) -> Optional[VPNConnection]:
        """Returns the current VPN connection or None if there isn't one."""
        return self.current_state.context.connection if self.current_state else None

    @property
    def current_server_id(self) -> Optional[str]:
        """
        Returns the server ID of the current VPN connection.

        Note that by if the current state is disconnected, `None` will be
        returned if a VPN connection was never established. Otherwise,
        the server ID of the last server the connection was established to
        will be returned instead.
        """
        return self.current_connection.server_id if self.current_connection else None

    @property
    def is_connection_active(self) -> bool:
        """Returns whether there is currently a VPN connection ongoing or not."""
        return not isinstance(self._current_state, (states.Disconnected, states.Error))

    @property
    def is_connected(self) -> bool:
        """Returns whether the user is connected to a VPN server or not."""
        return isinstance(self.current_state, states.Connected)

    @staticmethod
    def get_vpn_server(
            logical_server: LogicalServer, client_config: ClientConfig
    ) -> VPNServer:
        """
        :return: a :class:`proton.vpn.vpnconnection.interfaces.VPNServer` that
        can be used to establish a VPN connection with
        :class:`proton.vpn.vpnconnection.VPNConnection`.
        """
        physical_server = logical_server.get_random_physical_server()
        has_ipv6_support = ServerFeatureEnum.IPV6 in logical_server.features
        return VPNServer(
            server_ip=physical_server.entry_ip,
            domain=physical_server.domain,
            x25519pk=physical_server.x25519_pk,
            openvpn_ports=ProtocolPorts(
                udp=client_config.openvpn_ports.udp,
                tcp=client_config.openvpn_ports.tcp,
                tls=client_config.openvpn_ports.tls
            ),
            wireguard_ports=ProtocolPorts(
                udp=client_config.wireguard_ports.udp,
                tcp=client_config.wireguard_ports.tcp,
                tls=client_config.wireguard_ports.tls
            ),
            server_id=logical_server.id,
            server_name=logical_server.name,
            has_ipv6_support=has_ipv6_support,
            label=physical_server.label
        )

    def get_available_protocols_for_backend(
            self, backend_name: str
    ) -> Optional[PluggableComponent]:
        """Returns available protocols for the `backend_name`

        raises RuntimeError:  if no backends could be found."""
        backend_class = Loader.get("backend", class_name=backend_name)

        # This loader implementation is not ideal. We need to filter the
        # protocols by their validate methods.
        supported_protocols = [p for p in Loader.get_all(backend_class.backend)
                               if p.cls._validate()]  # pylint: disable=W0212

        return supported_protocols

    # pylint: disable=too-many-arguments
    async def connect(
            self, server: VPNServer,
            protocol: str = None,
            backend: str = None
    ):
        """Connects to a VPN server."""
        if not self._session_holder.session.logged_in:
            raise RuntimeError("Log in required before starting VPN connections.")

        logger.info(
            f"{server} / Protocol: {protocol} / Backend: {backend}",
            category="CONN", subcategory="CONNECT", event="START"
        )

        # Sets the settings to be applied when establishing the next connection.
        settings = await self.get_settings()
        # FIXME: this adds a big delay before creating the connection  # pylint: disable=fixme
        self._set_ks_setting(KillSwitchSetting(settings.killswitch), settings.protocol)
        self._set_split_tunneling_setting(settings.features.split_tunneling)

        protocol = protocol or settings.protocol

        connection = VPNConnection.create(
            server, self.credentials, settings, protocol, backend
        )

        connection.register(self._on_connection_event)

        await self._on_connection_event(
            events.Up(events.EventContext(connection=connection))
        )

    async def disconnect(self):
        """Disconnects the current VPN connection, if any."""
        await self._on_connection_event(
            events.Down(events.EventContext(connection=self.current_connection))
        )

    def register(self, subscriber: VPNStateSubscriber):
        """
        Registers a new subscriber to connection status updates.

        The subscriber should have a ```status_update``` method, which will
        be called passing it the new connection status whenever it changes.

        :param subscriber: Subscriber to register.
        """
        if not isinstance(subscriber, VPNStateSubscriber):
            raise ValueError(
                "The specified subscriber does not implement the "
                f"{VPNStateSubscriber.__name__} protocol."
            )
        self._publisher.register(subscriber.status_update)

    def unregister(self, subscriber: VPNStateSubscriber):
        """
        Unregister a subscriber from connection status updates.
        :param subscriber: Subscriber to unregister.
        """
        if not isinstance(subscriber, VPNStateSubscriber):
            raise ValueError(
                "The specified subscriber does not implement the "
                f"{VPNStateSubscriber.__name__} protocol."
            )
        self._publisher.unregister(subscriber.status_update)

    async def _handle_on_event(self, event: events.Event):
        """
        Handles the event by updating the current state of the connection,
        and returning a new event to be processed if any.
        """
        try:
            new_state = self.current_state.on_event(event)
        except FeatureSyntaxError as excp:
            self._usage_reporting.report_error(excp)
            logger.exception(msg=excp.message)
        except FeatureError as excp:
            logger.warning(msg=excp.message)
        except Exception as excp:
            self._usage_reporting.report_error(excp)
            raise excp
        else:
            return await self._update_state(new_state)
        return None

    async def _on_connection_event(self, event: events.Event):
        """
        Callback called when a connection event happens.
        """
        # The following lock guaranties that each new event is processed only
        # when the previous event was fully processed.
        async with self._lock:
            triggered_events = 0
            while event:
                triggered_events += 1
                if triggered_events > 99:
                    raise RuntimeError("Maximum number of chained connection events was reached.")
                event = await self._handle_on_event(event)

    async def _update_state(self, new_state) -> Optional[events.Event]:
        if new_state is self.current_state:
            return None

        old_state = self._current_state

        if isinstance(new_state.context.event, events.TwoFARequired) and \
           isinstance(old_state.context.event, events.TwoFARequired):
            return None

        self._current_state = new_state

        logger.info(
            f"{type(self._current_state).__name__}"
            f"{' (initial state)' if not old_state else ''}",
            category="CONN", event="STATE_CHANGED"
        )

        if isinstance(self._current_state, states.Disconnected) \
                and self._current_state.context.connection:
            # Unregister from connection event updates once the connection ended.
            self._current_state.context.connection.unregister(self._on_connection_event)

        if isinstance(old_state, states.Connected) and isinstance(new_state, states.Connected):
            # A Connected state can transition to a new Connected state when local agent
            # sends a new connected event, e.g. with a new port forwarding port. In this case,
            # the connection subscribers are notified (to update data on the client UI)
            # but the tasks associated with the connected state do not need to be run again
            # since the connection state didn't really change, only its context data did.
            self._publisher.notify(new_state)
            return None

        if self._current_state.notify_early:
            self._publisher.notify(new_state)
            new_event = await self._current_state.run_tasks()
        else:
            new_event = await self._current_state.run_tasks()
            self._publisher.notify(new_state)

        if (
            not self._current_state.context.reconnection
            and isinstance(self._current_state, states.Disconnected)
        ):
            self._set_ks_impl((await self.get_settings()).protocol)

        return new_event

    def _on_state_change_update_location(self, state: states.State):
        """Updates the user location when the connection is established."""
        connection_details = self._get_connection_details_from_state(state)
        if not connection_details:
            return

        current_location = self._session_holder.session.vpn_account.location

        self._session_holder.session.set_location(
            self._create_new_vpn_location(connection_details, current_location)
        )

    def _get_connection_details_from_state(
            self, state: states.State
    ) -> Optional[events.ConnectionDetails]:
        if not isinstance(state, states.Connected):
            return None

        connection_details = state.context.event.context.connection_details
        if not connection_details or not connection_details.device_ip:
            return None

        return connection_details

    def _create_new_vpn_location(self, connection_details, current_location) -> VPNLocation:
        return VPNLocation(
            IP=connection_details.device_ip,
            Country=connection_details.device_country,
            ISP=current_location.ISP,
            Long=current_location.Long,
            Lat=current_location.Lat
        )

    def _set_ks_impl(self, protocol: str):
        """
        By using this specific method we're leaking implementation details.

        Because we currently have to deal with two kill switch NetworkManager implementations,
        one for OpenVPN and one for WireGuard, and them not being compatible with each other,
        we need to ensure that when switching protocols,
        we only do this when we are in `Disconnected` state, to ensure
        that the environment is clean and we don't leave any residuals on a users machine.
        """
        kill_switch_backend = KillSwitch.get(protocol=protocol)
        StateContext.kill_switch = self._kill_switch or kill_switch_backend()

    def _set_split_tunneling_setting(self, st_setting: SplitTunnelingSetting):

        st_setting = deepcopy(st_setting)
        st_setting.enabled = (
            self._split_tunneling
            and st_setting.enabled
            and not self._is_free_tier()
        )

        StateContext.split_tunneling_setting = st_setting

    def _set_split_tunneling_impl(self):
        StateContext.split_tunneling = self._split_tunneling

    def _get_user_tier(self) -> int:
        # Default to free tier if session is not loaded yet
        return self._session_holder.user_tier or 0

    def _is_free_tier(self) -> bool:
        return self._get_user_tier() == 0

    def subscribe_to_certificate_updates(self, refresher: VPNDataRefresher):
        """Subscribes to certificate updates."""
        refresher.set_certificate_updated_callback(self._on_certificate_updated)

    async def _on_certificate_updated(self):
        """Actions to be taken when once the certificate is updated."""
        if isinstance(self.current_state, (states.Connected, states.Error)):
            await self.update_credentials()


class Subscriber:
    """
    Connection subscriber implementation that allows blocking until a certain state is reached.
    """
    def __init__(self):
        self.state: ConnectionStateEnum = None
        self.events = {state: threading.Event() for state in ConnectionStateEnum}

    def status_update(self, state):
        """
        This method will be called whenever a VPN connection state update occurs.
        :param state: new state.
        """
        self.state = state.type
        self.events[self.state].set()
        self.events[self.state].clear()

    def wait_for_state(self, state: ConnectionStateEnum, timeout: int = None):
        """
        Blocks until the specified VPN connection state is reached.

        :param state: target connection state.
        :param timeout: if specified, a TimeoutError will be raised
        when the target state is reached.
        """
        state_reached = self.events[state].wait(timeout)
        if not state_reached:
            raise TimeoutError(f"Time out occurred before reaching state {state.name}.")



"""
Wrapper over the NetworkManager client.


Copyright (c) 2023 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
# pylint: disable=duplicate-code

from concurrent.futures import Future
from threading import Thread, Lock
from typing import Optional

from packaging.version import Version

import gi
gi.require_version("NM", "1.0")
from gi.repository import NM, GLib, Gio, GObject  # pylint: disable=C0413 # noqa: E402

from proton.vpn import logging  # noqa: E402 pylint: disable=wrong-import-position

logger = logging.getLogger(__name__)


def _create_future():
    """Creates a future and sets its internal state as running."""
    future = Future()
    future.set_running_or_notify_cancel()
    return future


class NMClient:
    """
    Wrapper over the NetworkManager client.
    It also starts the GLib main loop used by the NetworkManager client.
    """
    _lock = Lock()
    _main_context = None
    _nm_client = None

    @classmethod
    def initialize_nm_client_singleton(cls):
        """
        Initializes the NetworkManager client singleton.

        If the singleton was initialized, this method will do nothing. However,
        if the singleton wasn't initialized it will initialize it, starting
        a new GLib MainLoop.

        A double-checked lock is used to avoid the possibility of multiple
        threads concurrently creating multiple instances of the NM client
        (with their own main loops).
        """
        if cls._nm_client:
            return

        with cls._lock:
            if not cls._nm_client:
                cls._initialize_nm_client_singleton()

    @classmethod
    def _initialize_nm_client_singleton(cls):
        cls._main_context = GLib.MainContext()

        # Setting daemon=True when creating the thread makes that this thread
        # exits abruptly when the python process exits. It would be better to
        # exit the thread running the main loop calling self._main_loop.quit().
        Thread(target=cls._run_glib_loop, daemon=True).start()

        def _init_nm_client():
            # It's important the NM.Client instance is created in the thread
            # running the GLib event loop so that then that's the thread used
            # for all GLib asynchronous operations.
            return NM.Client.new(cancellable=None)

        cls._nm_client = cls._run_on_glib_loop_thread(_init_nm_client).result()

    @classmethod
    def _run_glib_loop(cls):
        main_loop = GLib.MainLoop(cls._main_context)
        cls._main_context.push_thread_default()
        main_loop.run()

    @classmethod
    def _assert_running_on_glib_loop_thread(cls):
        """
        This method asserts that the thread running it is the one iterating
        GLib's main loop.

        It's useful to call this method at the beginning of any code block
        that's supposed to run in GLib's main loop, to avoid hard-to-debug
        issues.

        For more info:
        https://developer.gnome.org/documentation/tutorials/main-contexts.html#checking-threading
        """
        if not cls._main_context.is_owner():
            raise RuntimeError("Code being run outside GLib's main loop.")

    @classmethod
    def _run_on_glib_loop_thread(cls, function, *args, **kwargs) -> Future:
        future = _create_future()

        def wrapper():
            cls._assert_running_on_glib_loop_thread()
            try:
                future.set_result(function(*args, **kwargs))
            except BaseException as exc:  # pylint: disable=broad-except
                future.set_exception(exc)

        cls._main_context.invoke_full(priority=GLib.PRIORITY_DEFAULT, function=wrapper)

        return future

    def __init__(self):
        self.initialize_nm_client_singleton()

    def add_connection_async(
        self, connection: NM.Connection, save_to_disk: bool = False
    ) -> Future:
        """
        Adds a new connection asynchronously.
        https://lazka.github.io/pgi-docs/#NM-1.0/classes/Client.html#NM.Client.add_connection_async
        :param connection: connection to be added.
        :return: a Future to keep track of completion.
        """
        future_conn_activated = _create_future()

        def _on_interface_state_changed(_device, new_state, _old_state, _reason):
            """
            Monitors kill switch interface state changes and resolves
            the future as soon as the interface reaches the activated state
            """
            logger.debug(
                f"{connection.get_interface_name()} interface state changed "
                f"to {NM.DeviceState(new_state).value_name}"
            )
            if (
                    NM.DeviceState(new_state) == NM.DeviceState.ACTIVATED
                    and not future_conn_activated.done()
            ):
                future_conn_activated.set_result(None)

        def _on_interface_added(_nm_client, device):
            """
            Monitors interface creation. As soon as the kill switch interface
            is created it sets up the call back to monitor interface state changes.
            """
            logger.debug(
                f"{device.get_iface()} interface added in state {device.get_state().value_name}"
            )
            if not device.get_iface() == connection.get_interface_name():
                return

            handler_id = device.connect("state-changed", _on_interface_state_changed)
            future_conn_activated.add_done_callback(
                lambda f: self._run_on_glib_loop_thread(
                    GObject.signal_handler_disconnect, device, handler_id
                ).result()
            )

        def _on_connection_added(nm_client, res, _user_data):
            try:
                # Make sure exceptions creating the connection are passed to the future.
                nm_client.add_connection_finish(res)
            except Exception as exc:  # pylint: disable=broad-except
                future_conn_activated.set_exception(
                    RuntimeError(
                        f"Error setting adding KS connection: {nm_client=}, {res=}"
                    ).with_traceback(exc.__traceback__)
                )
                return

        def _add_connection_async():
            # Set up interface connection monitoring, which resolves the future
            # once the kill switch is active.
            handler_id = self._nm_client.connect("device-added", _on_interface_added)
            future_conn_activated.add_done_callback(
                lambda f: self._run_on_glib_loop_thread(
                    GObject.signal_handler_disconnect, self._nm_client, handler_id
                ).result()
            )

            # Add kill switch connection asynchronously.
            self._nm_client.add_connection_async(
                connection=connection,
                save_to_disk=save_to_disk,
                cancellable=None,
                callback=_on_connection_added,
                user_data=None
            )

        self._run_on_glib_loop_thread(_add_connection_async).result()

        return future_conn_activated

    def remove_connection_async(
            self, connection: NM.RemoteConnection
    ) -> Future:
        """
        Removes the specified connection asynchronously.
        https://lazka.github.io/pgi-docs/#NM-1.0/classes/RemoteConnection.html#NM.RemoteConnection.delete_async
        :param connection: connection to be removed.
        :return: a Future to keep track of completion.
        """
        future_interface_removed = _create_future()

        def _on_connection_removed(connection, result, _user_data):
            try:
                connection.delete_finish(result)
            except Exception as exc:  # pylint: disable=broad-except
                future_interface_removed.set_exception(
                    RuntimeError(
                        f"Error removing KS connection: {connection=}, {result=}"
                    ).with_traceback(exc.__traceback__)
                )

        def _on_interface_removed(_nm_client, device):
            logger.debug(
                f"{device.get_iface()} was removed."
            )
            if device.get_iface() == connection.get_interface_name():
                future_interface_removed.set_result(None)

        def _remove_connection_async():
            handler_id = self._nm_client.connect("device-removed", _on_interface_removed)
            future_interface_removed.add_done_callback(
                lambda f: self._run_on_glib_loop_thread(
                    GObject.signal_handler_disconnect, self._nm_client, handler_id
                ).result()
            )

            connection.delete_async(
                None,
                _on_connection_removed,
                None
            )

        self._run_on_glib_loop_thread(_remove_connection_async).result()

        return future_interface_removed

    def get_active_connection(self, conn_id: str) -> Optional[NM.ActiveConnection]:
        """
        Returns the specified active connection, if existing.
        :param conn_id: ID of the active connection.
        :return: the active connection if it was found. Otherwise, None.
        """
        def _get_active_connection():
            active_connections = self._nm_client.get_active_connections()

            for connection in active_connections:
                if connection.get_id() == conn_id:
                    return connection

            return None

        return self._run_on_glib_loop_thread(_get_active_connection).result()

    def get_connection(self, conn_id: str) -> Optional[NM.RemoteConnection]:
        """
        Returns the specified connection, if existing.
        :param conn_id: ID of the connection.
        :return: the connection if it was found. Otherwise, None.
        """
        return self._run_on_glib_loop_thread(
            self._nm_client.get_connection_by_id, conn_id
        ).result()

    def get_nm_running(self) -> bool:
        """Returns if NetworkManager daemon is running or not."""
        return self._run_on_glib_loop_thread(
            self._nm_client.get_nm_running
        ).result()

    def connectivity_check_get_enabled(self) -> bool:
        """Returns if connectivity check is enabled or not."""
        return self._run_on_glib_loop_thread(
            self._nm_client.connectivity_check_get_enabled
        ).result()

    def disable_connectivity_check(self) -> Future:
        """Since `connectivity_check_set_enabled` has been deprecated,
        we have to resort to lower lever commands.
        https://lazka.github.io/pgi-docs/#NM-1.0/classes/Client.html#NM.Client.connectivity_check_set_enabled

        This change is necessary since if this feature is enabled,
        dummy connection are inflated with a value of 20000.

        https://developer-old.gnome.org/NetworkManager/stable/NetworkManager.conf.html
        (see under `connectivity section`)
        """
        if Version(self._nm_client.get_version()) < Version("1.24.0"):
            # NM.Client.connectivity_check_set_enabled is deprecated since version 1.22
            # but the replacement method is only available in version 1.24.
            return self._run_on_glib_loop_thread(
                self._nm_client.connectivity_check_set_enabled, False
            )

        return self._dbus_set_property(
            object_path="/org/freedesktop/NetworkManager",
            interface_name="org.freedesktop.NetworkManager",
            property_name="ConnectivityCheckEnabled",
            value=GLib.Variant("b", False),
            timeout_msec=-1,
            cancellable=None
        )

    def _dbus_set_property(
            self, *userdata, object_path: str, interface_name: str, property_name: str,
            value: GLib.Variant, timeout_msec: int = -1,
            cancellable: Gio.Cancellable = None,
    ) -> Future:  # pylint: disable=too-many-arguments
        """Set NM properties since dedicated methods have been deprecated deprecated.
        Source: https://lazka.github.io/pgi-docs/#NM-1.0/classes/Client.html"""  # noqa

        future = _create_future()

        def _on_property_set(nm_client, res, _user_data):
            if not nm_client or not res or not nm_client.dbus_set_property_finish(res):
                future.set_exception(
                    RuntimeError(
                        f"Error disabling network connectivity check: {nm_client=}, {res=}"
                    )
                )
                return

            future.set_result(None)

        def _set_property_async():
            self._assert_running_on_glib_loop_thread()
            self._nm_client.dbus_set_property(
                object_path, interface_name, property_name,
                value, timeout_msec, cancellable, _on_property_set,
                userdata
            )

        self._run_on_glib_loop_thread(_set_property_async).result()

        return future



"""
List of exceptions raised in this package.


Copyright (c) 2023 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
from __future__ import annotations

from proton.session.exceptions import ProtonError


class ProtonVPNError(ProtonError):
    """Base exception for Proton VPN errors."""


class ServerNotFound(ProtonVPNError):
    """A VPN server was expected but was not found."""


class VPNDaemonError(Exception):
    """Base class for Proton API specific exceptions"""

    def __init__(self, message, additional_context=None):
        self.message = message
        self.additional_context = additional_context
        super().__init__(self.message)



"""
Cache Handler module.


Copyright (c) 2023 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""

import json
import os
from pathlib import Path
from proton.vpn import logging


logger = logging.getLogger(__name__)


class CacheHandler:
    """Used to save, load, and remove cache files."""
    def __init__(self, filepath: str):
        self._fp = Path(filepath)

    @property
    def exists(self):
        """True if the cache file exists and False otherwise."""
        return self._fp.is_file()

    def save(self, newdata: dict):
        """Save data to cache file."""
        self._fp.parent.mkdir(parents=True, exist_ok=True)
        with open(self._fp, "w", encoding="utf-8") as f:  # pylint: disable=C0103
            json.dump(newdata, f, indent=4)  # pylint: disable=C0103

    def load(self):
        """
        Load data from cache file, if it exists.
        If it exists, but content is not valid json, None is returned instead.
        """
        if not self.exists:
            return None

        try:
            with open(self._fp, "r", encoding="utf-8") as f:  # pylint: disable=C0103
                return json.load(f)  # pylint: disable=C0103
        except (json.decoder.JSONDecodeError, UnicodeDecodeError):
            filename = os.path.basename(self._fp)
            logger.warning(
                msg=f"Unable to decode JSON file \"{filename}\"",
                category="cache", event="load", exc_info=True
            )
            return None

    def remove(self):
        """ Remove cache from disk."""
        if self.exists:
            os.remove(self._fp)



"""
This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
# pylint: disable=duplicate-code

from proton.vpn import logging


logger = logging.getLogger(__name__)


def is_ipv6_disabled() -> bool:
    """Returns if IPv6 is disabled at kernel level or not."""
    filepath = "/sys/module/ipv6/parameters/disable"
    try:
        with open(filepath, "r", encoding="utf-8") as ipv6_disabled_file:
            if ipv6_disabled_file.read().strip() != "0":
                return True
    except FileNotFoundError:
        # If flatpak then we don't have access to the file and assume
        # IPv6 is enabled.
        logger.error("Unable to figure if IPv6 is enabled or disabled, probably flatpak install.")
        return False

    return False



"""
Copyright (c) 2023 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
import inspect
from typing import Optional, Callable
from datetime import timedelta
import random

from proton.vpn import logging
from proton.vpn.core.refresher.scheduler import RunAgain
from proton.vpn.core.session_holder import SessionHolder
from proton.vpn.session.credentials import VPNPubkeyCredentials
from proton.session.exceptions import (
    ProtonAPINotReachable, ProtonAPINotAvailable,
    ProtonAPIError
)

logger = logging.getLogger(__name__)

# pylint: disable=R0801


class CertificateRefresher:
    """
    Service in charge of refreshing certificate, that is used to derive
    users private keys, to establish VPN connections.
    """

    def __init__(self, session_holder: SessionHolder):
        self._session_holder = session_holder
        self._number_of_failed_refresh_attempts = 0
        self.certificate_updated_callback: Optional[Callable] = None

    @property
    def _session(self):
        return self._session_holder.session

    @property
    def initial_refresh_delay(self):
        """Returns the initial delay before the first refresh."""
        return self._session.vpn_account \
            .vpn_credentials \
            .pubkey_credentials \
            .remaining_time_to_next_refresh

    async def refresh(self) -> RunAgain:
        """Fetches the new certificate from the REST API."""
        try:
            certificate = await self._session.fetch_certificate()
            next_refresh_delay = certificate.remaining_time_to_next_refresh
            self._number_of_failed_refresh_attempts = 0
            await self._notify()
        except ProtonAPIError as error:
            if error.http_code != 429:
                raise

            logger.warning(f"Certificate refresh failed {error}")
            next_refresh_delay = self._get_next_refresh_delay()
            self._number_of_failed_refresh_attempts += 1
        except (ProtonAPINotReachable, ProtonAPINotAvailable) as error:
            logger.warning(f"Certificate refresh failed: {error}")
            next_refresh_delay = self._get_next_refresh_delay()
            self._number_of_failed_refresh_attempts += 1
        except Exception:
            logger.error(  # noqa: E501 # pylint: disable=line-too-long # nosemgrep: python.lang.best-practice.logging-error-without-handling.logging-error-without-handling
                "Certificate refresh failed unexpectedly."
                "Stopping certificate refresh."
            )
            raise

        logger_prefix = "Next"
        if self._number_of_failed_refresh_attempts:
            logger_prefix = f"Attempt {self._number_of_failed_refresh_attempts} for"

        logger.info(
            f"{logger_prefix} certificate refresh scheduled in "
            f"{timedelta(seconds=next_refresh_delay)}"
        )

        return RunAgain.after_seconds(next_refresh_delay)

    async def update_if_necessary(self):
        """Fetches a new certificate from the REST API
           if the current certificate has expired or will soon expire"""
        pubkey_credentials = self._session.vpn_account.vpn_credentials.pubkey_credentials
        if pubkey_credentials.remaining_time_to_next_refresh > 0:
            return  # too early to update the certificate

        try:
            await self._session.fetch_certificate()
            await self._notify()
        except ProtonAPIError as error:
            if error.http_code != 429:
                raise

            logger.warning(f"Certificate refresh failed {error}")
        except (ProtonAPINotReachable, ProtonAPINotAvailable) as error:
            logger.warning(f"Certificate refresh failed: {error}")
        except Exception:
            logger.error(
                "Certificate refresh failed unexpectedly."
                "Stopping certificate refresh."
            )
            raise

    def _get_next_refresh_delay(self):
        return min(
            generate_backoff_value(self._number_of_failed_refresh_attempts),
            VPNPubkeyCredentials.get_refresh_interval_in_seconds()
        )

    async def _notify(self):
        if self.certificate_updated_callback is None:
            return

        if inspect.iscoroutinefunction(self.certificate_updated_callback):
            await self.certificate_updated_callback()  # pylint: disable=not-callable
        else:
            raise ValueError(
                "Expected coroutine function but found "
                f"{type(self.certificate_updated_callback)}"
            )


def generate_backoff_value(
        number_of_failed_refresh_attempts: int, backoff_in_seconds: int = 1,
        random_component: float = None
) -> int:
    """Generate and return a backoff value for when API calls fail,
    so it can retry again without DDoS'ing the API."""
    random_component = random_component or _generate_random_component()
    return backoff_in_seconds * 2 ** number_of_failed_refresh_attempts * random_component


def _generate_random_component() -> int:
    """Generates random component between 1 - randones_percentage and 1 + randomness_percentage."""
    return 1 + VPNPubkeyCredentials.REFRESH_RANDOMNESS *\
        (2 * random.random() - 1)  # nosec B311 # noqa: E501 # pylint: disable=line-too-long # nosemgrep: gitlab.bandit.B311



"""
Copyright (c) 2023 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
import logging
import os
import hashlib
import getpass
import re

from proton.vpn.core.session_holder import (
    ClientTypeMetadata, DISTRIBUTION_VERSION, DISTRIBUTION_ID)
from proton.vpn.session.utils import get_desktop_environment

DSN = "https://9a5ea555a4dc48dbbb4cfa72bdbd0899@vpn-api.proton.me/core/v4/reports/sentry/25"
SSL_CERT_FILE = "SSL_CERT_FILE"
MACHINE_ID = "/etc/machine-id"
PROTON_VPN = "protonvpn"

# This is how we anonymise the data sent to us in sentry messages.
#
# The regular expressions will be matched against lower and upper case matches
# so the regular expression doesn't needs to include casing.
#
# {user} is a special key that will be replaced with the username of the
# current user. It makes for a more accurant regex match.
PRIVACY_REPLACEMENTS = [
    # Username in home directory
    (r'\/home\/{user}\/', r"/home/<HIDDEN>/"),
    # Email addresses
    (r'[A-Z0-9._%+-]+@[A-Z0-9.-]+\.[A-Z]{{2,}}', r"<HIDDEN>"),
    # GPG ID addresses
    (r'gpg: encrypted with.*', r"<HIDDEN>"),
]

log = logging.getLogger(__name__)


def scrub_private_data(data, keys):
    """
    Recursively scrub private data from any values in the event.

    :param data: The value that needs to be scrubbed.
    :param keys: A dictionary of values to substitute before evaluating the
        regex replacement.
    :returns: The contents of data with any discovered personal information
        replaced by <HIDDEN>.
    """
    if isinstance(data, (tuple, list)):
        for index, value in enumerate(data):
            data[index] = scrub_private_data(value, keys)
    elif isinstance(data, dict):
        for key, value in data.items():
            data[key] = scrub_private_data(value, keys)
    elif isinstance(data, str):
        for expression, replacement in PRIVACY_REPLACEMENTS:
            pattern = expression.format(**keys)
            data = re.sub(pattern, replacement, data, flags=re.IGNORECASE)
    return data


class UsageReporting:
    """Sends anonymous usage reports to Proton."""

    def __init__(self, client_type_metadata: ClientTypeMetadata):
        self._enabled = False
        self._capture_exception = None
        self._client_type_metadata = client_type_metadata
        self._user_id = None
        self._desktop_environment = get_desktop_environment()

    @property
    def enabled(self):
        """Returns whether anonymous usage reporting is enabled."""
        return self._enabled

    @enabled.setter
    def enabled(self, value: bool):
        """
        Sets whether usage reporting is enabled/disabled.

        On unsupported platforms, this may fail, in which case UsageReporting
        will be disabled and an exception will be logged.
        """

        try:
            self._enabled = value and self._start_sentry()
        except Exception:  # pylint: disable=broad-except
            self._enabled = False
            log.exception("Failed to enabled usage reporting")

    def report_error(self, error):
        """
        Send an error to sentry if anonymous usage reporting is enabled.

        On unsupported platforms, this may fail, in which case the error will
        will not be reported and an exception will be logged.
        """

        try:
            if self._enabled:
                self._add_scope_metadata()
                self._capture_exception(error)

        except Exception:  # pylint: disable=broad-except
            log.exception("Failed to report error '%s'", str(error))

    @staticmethod
    def _get_user_id(machine_id_filepath=MACHINE_ID, user_name=None):
        """
        Returns a unique identifier for the user.

        :param machine_id_filepath: The path to the machine id file,
            defaults to /etc/machine-id. This can be overrided for testing.

        :param user_name: The username to include in the hash, if None is
            provided, the current user is obtained from the environment.
        """

        if not os.path.exists(machine_id_filepath):
            return None

        # We include the username in the hash to avoid collisions on machines
        # with multiple users.
        if not user_name:
            user_name = getpass.getuser()

        # We use the machine id to uniquely identify the machine, we combine it
        # with the application name and the username. All three are hashed to
        # avoid leaking any personal information.
        with open(machine_id_filepath, "r", encoding="utf-8") as machine_id_file:
            machine_id = machine_id_file.read().strip()

        combined = hashlib.sha256(machine_id.encode('utf-8'))
        combined.update(hashlib.sha256(PROTON_VPN.encode('utf-8')).digest())
        combined.update(hashlib.sha256(user_name.encode('utf-8')).digest())

        return str(combined.hexdigest())

    @staticmethod
    def _sanitize_event(event, _hint, user_name=getpass.getuser()):
        """
        Sanitize the event before sending it to sentry.

        :param event: A dictionary representing the event to sanitize.
        :param _hint: Unused but required by the sentry SDK.
        :param user_name: The username to replace in the event, defaults to the
            current user, but can be set for testing purposes.
        """

        return scrub_private_data(
            event,
            {
                "user": user_name
            }
        )

    def _add_scope_metadata(self):
        """
        Unfortunately, we cannot set the user and tags on the isolation scope
        on startup because this is lost by the time we report an error.
        So we have to set the user and tags on the current scope just before
        reporting an error.
        """
        import sentry_sdk  # pylint: disable=import-outside-toplevel

        # Using configure_scope to set a tag works with older versions of
        # sentry (0.12.2) and so works on ubuntu 20.
        with sentry_sdk.configure_scope() as scope:
            scope.set_tag("distro_name", DISTRIBUTION_ID)
            scope.set_tag("distro_version", DISTRIBUTION_VERSION)
            scope.set_tag("desktop_environment", self._desktop_environment)
            if self._user_id and hasattr(scope, "set_user"):
                scope.set_user({"id": self._user_id})

    def _start_sentry(self):
        """Starts the sentry SDK with the appropriate configuration."""

        if self._capture_exception:
            return True

        if not self._client_type_metadata:
            raise ValueError("Client type metadata is not set, "
                             "UsageReporting.init() must be called first.")

        import sentry_sdk  # pylint: disable=import-outside-toplevel

        from sentry_sdk.integrations.dedupe import DedupeIntegration  # pylint: disable=import-outside-toplevel
        from sentry_sdk.integrations.stdlib import StdlibIntegration  # pylint: disable=import-outside-toplevel
        from sentry_sdk.integrations.modules import ModulesIntegration  # pylint: disable=import-outside-toplevel

        # Read from SSL_CERT_FILE from environment variable, this allows us to
        # use an http proxy if we want to.
        ca_certs = os.environ.get(SSL_CERT_FILE, None)
        client_type_metadata = self._client_type_metadata
        sentry_sdk.init(
            dsn=DSN,
            before_send=UsageReporting._sanitize_event,
            release=f"{client_type_metadata.type}-{client_type_metadata.version}",
            server_name=False,           # Don't send the computer name
            default_integrations=False,  # We want to be explicit about the integrations we use
            integrations=[
                DedupeIntegration(),     # Yes we want to avoid event duplication
                StdlibIntegration(),     # Yes we want info from the standard lib objects
                ModulesIntegration()     # Yes we want to know what python modules are installed
            ],
            ca_certs=ca_certs
        )

        # Store the user id so we don't have to calculate it again.
        self._user_id = self._get_user_id()

        # Store _capture_exception as a member, so it's easier to test.
        self._capture_exception = sentry_sdk.capture_exception

        return True
